"""UI view permission tests for the timeline app"""

from django.test import override_settings
from django.urls import reverse

# Projectroles dependency
from projectroles.tests.test_permissions import TestProjectPermissionBase


class TestTimelinePermissions(TestProjectPermissionBase):
    """Tests for timeline views"""

    def test_project_list(self):
        """Test project event list"""
        url = reverse(
            'timeline:list_project', kwargs={'project': self.project.sodar_uuid}
        )
        good_users = [
            self.superuser,
            self.owner_as.user,
            self.delegate_as.user,
            self.contributor_as.user,
            self.guest_as.user,
        ]
        bad_users = [self.user_no_roles, self.anonymous]
        self.assert_response(url, good_users, 200)
        self.assert_response(url, bad_users, 302)
        # Test public project
        self.project.set_public()
        self.assert_response(url, self.user_no_roles, 200)
        self.assert_response(url, self.anonymous, 302)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_project_list_anon(self):
        """Test project event list with anonynomus access"""
        url = reverse(
            'timeline:list_project', kwargs={'project': self.project.sodar_uuid}
        )
        self.project.set_public()
        self.assert_response(url, self.anonymous, 200)

    def test_site_list(self):
        """Test site event list"""
        url = reverse('timeline:list_site')
        good_users = [
            self.superuser,
            self.owner_as.user,
            self.delegate_as.user,
            self.contributor_as.user,
            self.guest_as.user,
            self.user_no_roles,
        ]
        self.assert_response(url, good_users, 200)
        self.assert_response(url, self.anonymous, 302)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_stite_list_anon(self):
        """Test site event list with anonynomus access"""
        url = reverse('timeline:list_site')
        self.assert_response(url, self.anonymous, 302)

    def test_project_event_object_list(self):
        """Test project event object list"""
        url = reverse(
            'timeline:list_object',
            kwargs={
                'project': self.project.sodar_uuid,
                'object_model': 'User',
                'object_uuid': self.user_owner.sodar_uuid,
            },
        )
        good_users = [
            self.superuser,
            self.owner_as.user,
            self.delegate_as.user,
            self.contributor_as.user,
            self.guest_as.user,
        ]
        bad_users = [self.user_no_roles, self.anonymous]
        self.assert_response(url, good_users, 200)
        self.assert_response(url, bad_users, 302)
        # Test public project
        self.project.set_public()
        self.assert_response(url, self.user_no_roles, 200)
        self.assert_response(url, self.anonymous, 302)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_project_event_object_list_anon(self):
        """Test object event list with anonynomus access"""
        url = reverse(
            'timeline:list_object',
            kwargs={
                'project': self.project.sodar_uuid,
                'object_model': 'User',
                'object_uuid': self.user_owner.sodar_uuid,
            },
        )
        self.project.set_public()
        self.assert_response(url, self.anonymous, 200)

    def test_site_event_object_list(self):
        """Test site event object list"""
        url = reverse(
            'timeline:list_object_site',
            kwargs={
                'object_model': 'User',
                'object_uuid': self.user_owner.sodar_uuid,
            },
        )
        good_users = [
            self.superuser,
            self.owner_as.user,
            self.delegate_as.user,
            self.contributor_as.user,
            self.guest_as.user,
            self.user_no_roles,
        ]
        self.assert_response(url, good_users, 200)
        self.assert_response(url, self.anonymous, 302)
        # Test public project
        self.project.set_public()
        self.assert_response(url, self.user_no_roles, 200)
        self.assert_response(url, self.anonymous, 302)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_site_event_object_list_anon(self):
        """Test site event list with anonymous access"""
        url = reverse(
            'timeline:list_object_site',
            kwargs={
                'object_model': 'User',
                'object_uuid': self.user_owner.sodar_uuid,
            },
        )
        self.project.set_public()
        self.assert_response(url, self.anonymous, 302)



"""Tests for management commands in the projectroles Django app"""

import os
from tempfile import NamedTemporaryFile

from django.conf import settings
from django.core import mail
from django.core.management import call_command
from django.test import override_settings
from djangoplugins.models import Plugin

from test_plus.test import TestCase

from projectroles.management.commands.cleanappsettings import START_MSG, END_MSG
from projectroles.management.commands.batchupdateroles import (
    Command as BatchUpdateRolesCommand,
)
from projectroles.models import (
    Role,
    RoleAssignment,
    ProjectInvite,
    AppSetting,
    SODAR_CONSTANTS,
)
from projectroles.tests.test_models import (
    ProjectMixin,
    RoleAssignmentMixin,
    ProjectInviteMixin,
    AppSettingMixin,
)

# SODAR constants
PROJECT_ROLE_OWNER = SODAR_CONSTANTS['PROJECT_ROLE_OWNER']
PROJECT_ROLE_DELEGATE = SODAR_CONSTANTS['PROJECT_ROLE_DELEGATE']
PROJECT_ROLE_CONTRIBUTOR = SODAR_CONSTANTS['PROJECT_ROLE_CONTRIBUTOR']
PROJECT_ROLE_GUEST = SODAR_CONSTANTS['PROJECT_ROLE_GUEST']
PROJECT_TYPE_CATEGORY = SODAR_CONSTANTS['PROJECT_TYPE_CATEGORY']
PROJECT_TYPE_PROJECT = SODAR_CONSTANTS['PROJECT_TYPE_PROJECT']

EXAMPLE_APP_NAME = 'example_project_app'
CLEAN_LOG_PREFIX = 'INFO:projectroles.management.commands.cleanappsettings:'


class BatchUpdateRolesMixin:
    """Helpers for batchupdateroles testing"""

    file = None

    def _write_file(self, data):
        """Write data to temporary CSV file"""
        if not isinstance(data[0], list):
            data = [data]
        self.file.write(
            bytes('\n'.join([';'.join(r) for r in data]), encoding='utf-8')
        )
        self.file.close()


class TestBatchUpdateRoles(
    ProjectMixin,
    RoleAssignmentMixin,
    ProjectInviteMixin,
    BatchUpdateRolesMixin,
    TestCase,
):
    """Tests for batchupdateroles command"""

    def setUp(self):
        super().setUp()

        # Init roles
        self.role_owner = Role.objects.get_or_create(name=PROJECT_ROLE_OWNER)[0]
        self.role_delegate = Role.objects.get_or_create(
            name=PROJECT_ROLE_DELEGATE
        )[0]
        self.role_contributor = Role.objects.get_or_create(
            name=PROJECT_ROLE_CONTRIBUTOR
        )[0]
        self.role_guest = Role.objects.get_or_create(name=PROJECT_ROLE_GUEST)[0]

        # Init users
        self.user_owner = self.make_user('owner')
        self.user_owner.email = 'owner_user@example.com'
        self.user_owner.save()
        self.user_owner_cat = self.make_user('owner_cat')
        self.user_owner_cat.email = 'cat_owner_user@example.com'
        self.user_owner_cat.save()

        # Init projects
        self.category = self._make_project(
            'top_category', PROJECT_TYPE_CATEGORY, None
        )
        self.cat_owner_as = self._make_assignment(
            self.category, self.user_owner_cat, self.role_owner
        )

        self.project = self._make_project(
            'sub_project', PROJECT_TYPE_PROJECT, self.category
        )
        self.owner_as = self._make_assignment(
            self.project, self.user_owner, self.role_owner
        )

        # Init command class
        self.command = BatchUpdateRolesCommand()

        # Init file
        self.file = NamedTemporaryFile(delete=False)

    def tearDown(self):
        if self.file:
            os.remove(self.file.name)
        super().tearDown()

    def test_invite(self):
        """Test inviting a single user via email to project"""
        p_uuid = str(self.project.sodar_uuid)
        email = 'new@example.com'

        # Assert preconditions
        self.assertEqual(ProjectInvite.objects.count(), 0)

        self._write_file([p_uuid, email, PROJECT_ROLE_GUEST])
        self.command.handle(
            **{'file': self.file.name, 'issuer': self.user_owner.username}
        )

        # Assert postconditions
        self.assertEqual(ProjectInvite.objects.count(), 1)
        invite = ProjectInvite.objects.first()
        self.assertEqual(invite.email, email)
        self.assertEqual(invite.project, self.project)
        self.assertEqual(invite.role, self.role_guest)
        self.assertEqual(invite.issuer, self.user_owner)
        self.assertEqual(len(mail.outbox), 1)

    def test_invite_existing(self):
        """Test inviting a user when they already have an active invite"""
        p_uuid = str(self.project.sodar_uuid)
        email = 'new@example.com'
        self._make_invite(email, self.project, self.role_guest, self.user_owner)

        # Assert preconditions
        self.assertEqual(ProjectInvite.objects.count(), 1)

        self._write_file([p_uuid, email, PROJECT_ROLE_GUEST])
        self.command.handle(
            **{'file': self.file.name, 'issuer': self.user_owner.username}
        )

        # Assert postconditions
        self.assertEqual(ProjectInvite.objects.count(), 1)
        self.assertEqual(len(mail.outbox), 0)

    def test_invite_multi_user(self):
        """Test inviting multiple users"""
        p_uuid = str(self.project.sodar_uuid)
        email = 'new@example.com'
        email2 = 'new2@example.com'

        # Assert preconditions
        self.assertEqual(ProjectInvite.objects.count(), 0)

        fd = [
            [p_uuid, email, PROJECT_ROLE_GUEST],
            [p_uuid, email2, PROJECT_ROLE_GUEST],
        ]
        self._write_file(fd)
        self.command.handle(
            **{'file': self.file.name, 'issuer': self.user_owner.username}
        )

        # Assert postconditions
        self.assertEqual(ProjectInvite.objects.count(), 2)
        self.assertEqual(len(mail.outbox), 2)

    def test_invite_multi_project(self):
        """Test inviting user to multiple projects"""
        c_uuid = str(self.category.sodar_uuid)
        p_uuid = str(self.project.sodar_uuid)
        email = 'new@example.com'

        # Assert preconditions
        self.assertEqual(ProjectInvite.objects.count(), 0)

        fd = [
            [c_uuid, email, PROJECT_ROLE_GUEST],
            [p_uuid, email, PROJECT_ROLE_GUEST],
        ]
        self._write_file(fd)
        # NOTE: Using user_owner_cat as they have perms for both projects
        self.command.handle(
            **{'file': self.file.name, 'issuer': self.user_owner_cat.username}
        )

        # Assert postconditions
        self.assertEqual(
            ProjectInvite.objects.filter(project=self.category).count(), 1
        )
        self.assertEqual(
            ProjectInvite.objects.filter(project=self.project).count(), 1
        )
        self.assertEqual(len(mail.outbox), 2)

    def test_invite_owner(self):
        """Test inviting an owner (should fail)"""
        p_uuid = str(self.project.sodar_uuid)
        email = 'new@example.com'
        self._write_file([p_uuid, email, PROJECT_ROLE_OWNER])
        self.command.handle(
            **{'file': self.file.name, 'issuer': self.user_owner.username}
        )

        # Assert postconditions
        self.assertEqual(ProjectInvite.objects.count(), 0)
        self.assertEqual(len(mail.outbox), 0)

    def test_invite_delegate(self):
        """Test inviting a delegate"""
        p_uuid = str(self.project.sodar_uuid)
        email = 'new@example.com'
        self._write_file([p_uuid, email, PROJECT_ROLE_DELEGATE])
        self.command.handle(
            **{'file': self.file.name, 'issuer': self.user_owner.username}
        )

        # Assert postconditions
        self.assertEqual(ProjectInvite.objects.count(), 1)
        self.assertEqual(len(mail.outbox), 1)

    @override_settings(PROJECTROLES_DELEGATE_LIMIT=2)
    def test_invite_delegate_no_perms(self):
        """Test inviting a delegate without perms (should fail)"""
        p_uuid = str(self.project.sodar_uuid)
        email = 'new@example.com'
        user_delegate = self.make_user('delegate')
        self._make_assignment(self.project, user_delegate, self.role_delegate)
        self._write_file([p_uuid, email, PROJECT_ROLE_DELEGATE])
        self.command.handle(
            **{'file': self.file.name, 'issuer': user_delegate.username}
        )

        # Assert postconditions
        self.assertEqual(ProjectInvite.objects.count(), 0)
        self.assertEqual(len(mail.outbox), 0)

    def test_invite_delegate_limit(self):
        """Test inviting a delegate with limit reached (should fail)"""
        p_uuid = str(self.project.sodar_uuid)
        email = 'new@example.com'
        user_delegate = self.make_user('delegate')
        self._make_assignment(self.project, user_delegate, self.role_delegate)

        self._write_file([p_uuid, email, PROJECT_ROLE_DELEGATE])
        self.command.handle(
            **{'file': self.file.name, 'issuer': self.user_owner.username}
        )

        # Assert postconditions
        self.assertEqual(ProjectInvite.objects.count(), 0)
        self.assertEqual(len(mail.outbox), 0)

    def test_role_add(self):
        """Test adding role to user already in system"""
        p_uuid = str(self.project.sodar_uuid)
        email = 'new@example.com'
        user_new = self.make_user('user_new')
        user_new.email = email
        user_new.save()

        # Assert preconditions
        self.assertEqual(ProjectInvite.objects.count(), 0)

        self._write_file([p_uuid, email, PROJECT_ROLE_GUEST])
        self.command.handle(
            **{'file': self.file.name, 'issuer': self.user_owner.username}
        )

        # Assert postconditions
        self.assertEqual(ProjectInvite.objects.count(), 0)
        role_as = RoleAssignment.objects.get(
            project=self.project, user=user_new
        )
        self.assertEqual(role_as.role, self.role_guest)
        self.assertEqual(len(mail.outbox), 1)

    def test_role_update(self):
        """Test updating an existing role for user"""
        p_uuid = str(self.project.sodar_uuid)
        email = 'new@example.com'
        user_new = self.make_user('user_new')
        user_new.email = email
        user_new.save()
        role_as = self._make_assignment(self.project, user_new, self.role_guest)

        # Assert preconditions
        self.assertEqual(
            RoleAssignment.objects.filter(
                project=self.project, user=user_new
            ).count(),
            1,
        )

        self._write_file([p_uuid, email, PROJECT_ROLE_CONTRIBUTOR])
        self.command.handle(
            **{'file': self.file.name, 'issuer': self.user_owner.username}
        )

        # Assert postconditions
        self.assertEqual(ProjectInvite.objects.count(), 0)
        role_as.refresh_from_db()
        self.assertEqual(role_as.role, self.role_contributor)
        self.assertEqual(len(mail.outbox), 1)

    def test_role_update_owner(self):
        """Test updating the role of current owner (should fail)"""
        p_uuid = str(self.project.sodar_uuid)
        email = 'owner@example.com'
        self.user_owner.email = email
        self.user_owner.save()

        self._write_file([p_uuid, email, PROJECT_ROLE_CONTRIBUTOR])
        self.command.handle(
            **{'file': self.file.name, 'issuer': self.user_owner.username}
        )

        # Assert postconditions
        self.assertEqual(ProjectInvite.objects.count(), 0)
        self.owner_as.refresh_from_db()
        self.assertEqual(self.owner_as.role, self.role_owner)
        self.assertEqual(
            RoleAssignment.objects.filter(
                project=self.project, user=self.user_owner
            ).count(),
            1,
        )
        self.assertEqual(len(mail.outbox), 0)

    @override_settings(PROJECTROLES_DELEGATE_LIMIT=2)
    def test_role_update_delegate_no_perms(self):
        """Test updating a delegate role without perms (should fail)"""
        p_uuid = str(self.project.sodar_uuid)
        email = 'new@example.com'
        user_new = self.make_user('user_new')
        user_new.email = email
        user_new.save()
        role_as = self._make_assignment(self.project, user_new, self.role_guest)
        user_delegate = self.make_user('delegate')
        self._make_assignment(self.project, user_delegate, self.role_delegate)

        self._write_file([p_uuid, email, PROJECT_ROLE_DELEGATE])
        self.command.handle(
            **{'file': self.file.name, 'issuer': user_delegate.username}
        )

        # Assert postconditions
        self.assertEqual(ProjectInvite.objects.count(), 0)
        role_as.refresh_from_db()
        self.assertEqual(role_as.role, self.role_guest)
        self.assertEqual(len(mail.outbox), 0)

    def test_role_update_delegate_limit(self):
        """Test updating a delegate role with limit reached (should fail)"""
        p_uuid = str(self.project.sodar_uuid)
        email = 'new@example.com'
        user_new = self.make_user('user_new')
        user_new.email = email
        user_new.save()
        role_as = self._make_assignment(self.project, user_new, self.role_guest)
        user_delegate = self.make_user('delegate')
        self._make_assignment(self.project, user_delegate, self.role_delegate)

        self._write_file([p_uuid, email, PROJECT_ROLE_DELEGATE])
        self.command.handle(
            **{'file': self.file.name, 'issuer': self.user_owner.username}
        )

        # Assert postconditions
        self.assertEqual(ProjectInvite.objects.count(), 0)
        role_as.refresh_from_db()
        self.assertEqual(role_as.role, self.role_guest)
        self.assertEqual(len(mail.outbox), 0)

    def test_role_add_inherited_owner(self):
        """Test adding a role of an inherited owner (should fail)"""
        p_uuid = str(self.project.sodar_uuid)
        email = self.user_owner_cat.email

        self._write_file([p_uuid, email, PROJECT_ROLE_CONTRIBUTOR])
        self.command.handle(
            **{'file': self.file.name, 'issuer': self.user_owner.username}
        )

        # Assert postconditions
        self.assertEqual(ProjectInvite.objects.count(), 0)
        self.assertEqual(
            RoleAssignment.objects.filter(
                project=self.project, user=self.user_owner_cat
            ).count(),
            0,
        )
        self.assertEqual(len(mail.outbox), 0)

    def test_invite_and_update(self):
        """Test inviting and updating in one command"""
        p_uuid = str(self.project.sodar_uuid)
        email = 'new@example.com'
        user_new = self.make_user('user_new')
        user_new.email = email
        user_new.save()
        email2 = 'new2@example.com'

        # Assert preconditions
        self.assertEqual(ProjectInvite.objects.count(), 0)

        fd = [
            [p_uuid, email, PROJECT_ROLE_GUEST],
            [p_uuid, email2, PROJECT_ROLE_GUEST],
        ]
        self._write_file(fd)
        self.command.handle(
            **{'file': self.file.name, 'issuer': self.user_owner.username}
        )

        # Assert postconditions
        self.assertEqual(ProjectInvite.objects.count(), 1)
        self.assertIsNotNone(
            RoleAssignment.objects.filter(
                project=self.project, user=user_new
            ).count(),
            1,
        )
        self.assertEqual(len(mail.outbox), 2)

    def test_command_no_issuer(self):
        """Test invite without issuer (should go under admin)"""
        admin = self.make_user(settings.PROJECTROLES_DEFAULT_ADMIN)
        admin.is_superuser = True
        admin.save()
        p_uuid = str(self.project.sodar_uuid)
        email = 'new@example.com'

        self._write_file([p_uuid, email, PROJECT_ROLE_GUEST])
        self.command.handle(**{'file': self.file.name, 'issuer': None})

        # Assert postconditions
        invite = ProjectInvite.objects.first()
        self.assertEqual(invite.issuer, admin)

    def test_command_no_perms(self):
        """Test invite with issuer who lacks permissions for a project"""
        issuer = self.make_user('issuer')
        p_uuid = str(self.project.sodar_uuid)
        email = 'new@example.com'

        self._write_file([p_uuid, email, PROJECT_ROLE_GUEST])
        self.command.handle(
            **{'file': self.file.name, 'issuer': issuer.username}
        )

        # Assert postconditions
        self.assertEqual(ProjectInvite.objects.count(), 0)
        self.assertEqual(len(mail.outbox), 0)


class TestCleanAppSettings(
    ProjectMixin, RoleAssignmentMixin, AppSettingMixin, TestCase
):
    """Tests for cleanappsettings command and associated functions"""

    def _make_undefined_setting(self):
        """Create database setting not reflected in the AppSetting dict"""
        ghost = AppSetting(
            app_plugin=self.plugin,
            project=self.project,
            name='ghost',
            type='BOOLEAN',
            value=True,
        )
        ghost.save()
        return ghost

    def setUp(self):
        super().setUp()

        # Init roles
        self.role_owner = Role.objects.get_or_create(name=PROJECT_ROLE_OWNER)[0]
        self.role_delegate = Role.objects.get_or_create(
            name=PROJECT_ROLE_DELEGATE
        )[0]
        self.role_contributor = Role.objects.get_or_create(
            name=PROJECT_ROLE_CONTRIBUTOR
        )[0]
        self.role_guest = Role.objects.get_or_create(name=PROJECT_ROLE_GUEST)[0]

        # Init users
        self.user_owner = self.make_user('owner')
        self.user_owner.email = 'owner_user@example.com'
        self.user_owner.save()

        # Init projects
        self.category = self._make_project(
            'top_category', PROJECT_TYPE_CATEGORY, None
        )
        self.cat_owner_as = self._make_assignment(
            self.category, self.user_owner, self.role_owner
        )

        self.project = self._make_project(
            'sub_project', PROJECT_TYPE_PROJECT, self.category
        )
        self.owner_as = self._make_assignment(
            self.project, self.user_owner, self.role_owner
        )
        self.plugin = Plugin.objects.get(name='example_project_app')

        # Init test setting
        self.setting_str_values = {
            'app_name': EXAMPLE_APP_NAME,
            'project': self.project,
            'name': 'project_str_setting',
            'setting_type': 'STRING',
            'value': 'test',
            'update_value': 'better test',
            'non_valid_value': False,
        }
        self.setting_int_values = {
            'app_name': EXAMPLE_APP_NAME,
            'project': self.project,
            'name': 'project_int_setting',
            'setting_type': 'INTEGER',
            'value': 0,
            'update_value': 170,
            'non_valid_value': 'Nan',
        }
        self.setting_bool_values = {
            'app_name': EXAMPLE_APP_NAME,
            'project': self.project,
            'name': 'project_bool_setting',
            'setting_type': 'BOOLEAN',
            'value': False,
            'update_value': True,
            'non_valid_value': 170,
        }
        self.setting_json_values = {
            'app_name': EXAMPLE_APP_NAME,
            'project': self.project,
            'name': 'project_json_setting',
            'setting_type': 'JSON',
            'value': {
                'Example': 'Value',
                'list': [1, 2, 3, 4, 5],
                'level_6': False,
            },
            'update_value': {'Test_more': 'often_always'},
            'non_valid_value': self.project,
        }
        self.setting_iprestrict_values = {
            'app_name': 'projectroles',
            'project': self.project,
            'name': 'ip_restrict',
            'setting_type': 'BOOLEAN',
            'value': False,
            'update_value': True,
            'non_valid_value': 170,
        }
        self.settings = [
            self.setting_int_values,
            self.setting_json_values,
            self.setting_str_values,
            self.setting_bool_values,
            self.setting_iprestrict_values,
        ]
        for s in self.settings:
            self._make_setting(
                app_name=s['app_name'],
                name=s['name'],
                setting_type=s['setting_type'],
                value=s['value'] if s['setting_type'] != 'JSON' else '',
                value_json=s['value'] if s['setting_type'] == 'JSON' else {},
                project=s['project'],
            )

    def test_command_cleanappsettings(self):
        """Test the cleanappsetting commmand"""
        self._make_undefined_setting()
        self.assertEqual(AppSetting.objects.count(), 6)

        with self.assertLogs(
            'projectroles.management.commands.cleanappsettings', level='INFO'
        ) as cm:
            call_command('cleanappsettings')
            self.assertEqual(
                cm.output,
                [
                    CLEAN_LOG_PREFIX + START_MSG,
                    (
                        CLEAN_LOG_PREFIX
                        + 'Deleting "settings.example_project_app.ghost" '
                        'from project "{}"'.format(self.project.title)
                    ),
                    CLEAN_LOG_PREFIX + END_MSG,
                ],
            )
        self.assertEqual(AppSetting.objects.count(), 5)



"""UI views for the appalerts app"""

from django.contrib import messages
from django.shortcuts import redirect
from django.urls import reverse
from django.views.generic import ListView, View

# Projectroles dependency
from projectroles.views import (
    LoginRequiredMixin,
    LoggedInPermissionMixin,
)

from appalerts.models import AppAlert


class AppAlertListView(LoginRequiredMixin, LoggedInPermissionMixin, ListView):
    """App alert list view"""

    permission_required = 'appalerts.view_alerts'
    template_name = 'appalerts/alert_list.html'

    def get_queryset(self):
        return AppAlert.objects.filter(
            user=self.request.user, active=True
        ).order_by('-pk')


class AppAlertLinkRedirectView(
    LoginRequiredMixin, LoggedInPermissionMixin, View
):
    """View for redirecting to alert link and dismissing alert"""

    permission_required = 'appalerts.view_alerts'

    def _handle_error(self, msg):
        messages.error(self.request, msg)
        return redirect(reverse('appalerts:list'))

    def get(self, request, **kwargs):
        alert = AppAlert.objects.filter(
            sodar_uuid=kwargs.get('appalert')
        ).first()

        # Handle errors
        if not alert:
            return self._handle_error('Alert not found.')
        if request.user != alert.user:
            return self._handle_error('Alert assigned to different user.')
        if not alert.url:
            return self._handle_error('No URL found for alert.')

        # All OK = dismiss and redirect
        alert.active = False
        alert.save()
        return redirect(alert.url)



"""Permission tests for the siteinfo app"""

from django.test import override_settings
from django.urls import reverse

# Projectroles dependency
from projectroles.tests.test_permissions import TestPermissionBase


class TestSiteInfoPermissions(TestPermissionBase):
    """Tests for siteinfo view permissions"""

    def setUp(self):
        # Create users
        self.superuser = self.make_user('superuser')
        self.superuser.is_superuser = True
        self.superuser.is_staff = True
        self.superuser.save()
        self.regular_user = self.make_user('regular_user')
        # No user
        self.anonymous = None

    def test_site_info(self):
        """Test site info view"""
        url = reverse('siteinfo:info')
        good_users = [self.superuser]
        bad_users = [self.anonymous, self.regular_user]
        self.assert_response(url, good_users, 200)
        self.assert_response(url, bad_users, 302)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_site_info_anon(self):
        """Test site info view with anonymous access"""
        url = reverse('siteinfo:info')
        self.assert_response(url, self.anonymous, 302)



from django.conf.urls import url

from projectroles import views, views_ajax, views_api, views_taskflow

app_name = 'projectroles'

# UI views
urls_ui = [
    url(
        regex=r'^(?P<project>[0-9a-f-]+)$',
        view=views.ProjectDetailView.as_view(),
        name='detail',
    ),
    url(
        regex=r'^update/(?P<project>[0-9a-f-]+)$',
        view=views.ProjectUpdateView.as_view(),
        name='update',
    ),
    url(
        regex=r'^create$', view=views.ProjectCreateView.as_view(), name='create'
    ),
    url(
        regex=r'^create/(?P<project>[0-9a-f-]+)$',
        view=views.ProjectCreateView.as_view(),
        name='create',
    ),
    # Search views
    url(
        regex=r'^search/results/$',
        view=views.ProjectSearchResultsView.as_view(),
        name='search',
    ),
    url(
        regex=r'^search/advanced$',
        view=views.ProjectAdvancedSearchView.as_view(),
        name='search_advanced',
    ),
    # Project role views
    url(
        regex=r'^members/(?P<project>[0-9a-f-]+)$',
        view=views.ProjectRoleView.as_view(),
        name='roles',
    ),
    url(
        regex=r'^members/create/(?P<project>[0-9a-f-]+)$',
        view=views.RoleAssignmentCreateView.as_view(),
        name='role_create',
    ),
    url(
        regex=r'^members/update/(?P<roleassignment>[0-9a-f-]+)$',
        view=views.RoleAssignmentUpdateView.as_view(),
        name='role_update',
    ),
    url(
        regex=r'^members/delete/(?P<roleassignment>[0-9a-f-]+)$',
        view=views.RoleAssignmentDeleteView.as_view(),
        name='role_delete',
    ),
    url(
        regex=r'^members/owner/transfer/(?P<project>[0-9a-f-]+)$',
        view=views.RoleAssignmentOwnerTransferView.as_view(),
        name='role_owner_transfer',
    ),
    # Project invite views
    url(
        regex=r'^invites/(?P<project>[0-9a-f-]+)$',
        view=views.ProjectInviteView.as_view(),
        name='invites',
    ),
    url(
        regex=r'^invites/create/(?P<project>[0-9a-f-]+)',
        view=views.ProjectInviteCreateView.as_view(),
        name='invite_create',
    ),
    url(
        regex=r'^invites/accept/(?P<secret>[\w\-]+)$',
        view=views.ProjectInviteAcceptView.as_view(),
        name='invite_accept',
    ),
    url(
        regex=r'^invites/process/ldap/(?P<secret>[\w\-]+)$',
        view=views.ProjectInviteProcessLDAPView.as_view(),
        name='invite_process_ldap',
    ),
    url(
        regex=r'^invites/process/local/(?P<secret>[\w\-]+)$',
        view=views.ProjectInviteProcessLocalView.as_view(),
        name='invite_process_local',
    ),
    url(
        regex=r'^invites/resend/(?P<projectinvite>[0-9a-f-]+)$',
        view=views.ProjectInviteResendView.as_view(),
        name='invite_resend',
    ),
    url(
        regex=r'^invites/revoke/(?P<projectinvite>[0-9a-f-]+)$',
        view=views.ProjectInviteRevokeView.as_view(),
        name='invite_revoke',
    ),
    url(
        regex=r'^user/update$',
        view=views.UserUpdateView.as_view(),
        name='user_update',
    ),
    # Remote site and project views
    url(
        regex=r'^remote/sites$',
        view=views.RemoteSiteListView.as_view(),
        name='remote_sites',
    ),
    url(
        regex=r'^remote/site/add$',
        view=views.RemoteSiteCreateView.as_view(),
        name='remote_site_create',
    ),
    url(
        regex=r'^remote/site/update/(?P<remotesite>[0-9a-f-]+)$',
        view=views.RemoteSiteUpdateView.as_view(),
        name='remote_site_update',
    ),
    url(
        regex=r'^remote/site/delete/(?P<remotesite>[0-9a-f-]+)$',
        view=views.RemoteSiteDeleteView.as_view(),
        name='remote_site_delete',
    ),
    url(
        regex=r'^remote/site/(?P<remotesite>[0-9a-f-]+)$',
        view=views.RemoteProjectListView.as_view(),
        name='remote_projects',
    ),
    url(
        regex=r'^remote/site/access/(?P<remotesite>[0-9a-f-]+)$',
        view=views.RemoteProjectBatchUpdateView.as_view(),
        name='remote_projects_update',
    ),
    url(
        regex=r'^remote/site/sync/(?P<remotesite>[0-9a-f-]+)$',
        view=views.RemoteProjectSyncView.as_view(),
        name='remote_projects_sync',
    ),
]

# Ajax API views
urls_ajax = [
    url(
        regex=r'^ajax/list$',
        view=views_ajax.ProjectListAjaxView.as_view(),
        name='ajax_project_list',
    ),
    url(
        regex=r'^ajax/list/columns$',
        view=views_ajax.ProjectListColumnAjaxView.as_view(),
        name='ajax_project_list_columns',
    ),
    url(
        regex=r'^ajax/list/roles',
        view=views_ajax.ProjectListRoleAjaxView.as_view(),
        name='ajax_project_list_roles',
    ),
    url(
        regex=r'^ajax/star/(?P<project>[0-9a-f-]+)',
        view=views_ajax.ProjectStarringAjaxView.as_view(),
        name='ajax_star',
    ),
    url(
        r'^ajax/autocomplete/user$',
        view=views_ajax.UserAutocompleteAjaxView.as_view(),
        name='ajax_autocomplete_user',
    ),
    url(
        r'^ajax/autocomplete/user/redirect$',
        view=views_ajax.UserAutocompleteRedirectAjaxView.as_view(
            create_field='user'
        ),
        name='ajax_autocomplete_user_redirect',
    ),
]

# REST API views
urls_api = [
    url(
        regex=r'^api/list$',
        view=views_api.ProjectListAPIView.as_view(),
        name='api_project_list',
    ),
    url(
        regex=r'^api/retrieve/(?P<project>[0-9a-f-]+)$',
        view=views_api.ProjectRetrieveAPIView.as_view(),
        name='api_project_retrieve',
    ),
    url(
        regex=r'^api/create$',
        view=views_api.ProjectCreateAPIView.as_view(),
        name='api_project_create',
    ),
    url(
        regex=r'^api/update/(?P<project>[0-9a-f-]+)$',
        view=views_api.ProjectUpdateAPIView.as_view(),
        name='api_project_update',
    ),
    url(
        regex=r'^api/roles/create/(?P<project>[0-9a-f-]+)$',
        view=views_api.RoleAssignmentCreateAPIView.as_view(),
        name='api_role_create',
    ),
    url(
        regex=r'^api/roles/update/(?P<roleassignment>[0-9a-f-]+)$',
        view=views_api.RoleAssignmentUpdateAPIView.as_view(),
        name='api_role_update',
    ),
    url(
        regex=r'^api/roles/destroy/(?P<roleassignment>[0-9a-f-]+)$',
        view=views_api.RoleAssignmentDestroyAPIView.as_view(),
        name='api_role_destroy',
    ),
    url(
        regex=r'^api/roles/owner-transfer/(?P<project>[0-9a-f-]+)$',
        view=views_api.RoleAssignmentOwnerTransferAPIView.as_view(),
        name='api_role_owner_transfer',
    ),
    url(
        regex=r'^api/invites/list/(?P<project>[0-9a-f-]+)$',
        view=views_api.ProjectInviteListAPIView.as_view(),
        name='api_invite_list',
    ),
    url(
        regex=r'^api/invites/create/(?P<project>[0-9a-f-]+)$',
        view=views_api.ProjectInviteCreateAPIView.as_view(),
        name='api_invite_create',
    ),
    url(
        regex=r'^api/invites/revoke/(?P<projectinvite>[0-9a-f-]+)$',
        view=views_api.ProjectInviteRevokeAPIView.as_view(),
        name='api_invite_revoke',
    ),
    url(
        regex=r'^api/invites/resend/(?P<projectinvite>[0-9a-f-]+)$',
        view=views_api.ProjectInviteResendAPIView.as_view(),
        name='api_invite_resend',
    ),
    url(
        regex=r'^api/users/list$',
        view=views_api.UserListAPIView.as_view(),
        name='api_user_list',
    ),
    url(
        regex=r'^api/users/current$',
        view=views_api.CurrentUserRetrieveAPIView.as_view(),
        name='api_user_current',
    ),
    url(
        regex=r'^api/remote/get/(?P<secret>[\w\-]+)$',
        view=views_api.RemoteProjectGetAPIView.as_view(),
        name='api_remote_get',
    ),
]

# Taskflow API views
urls_taskflow = [
    url(
        regex=r'^taskflow/get$',
        view=views_taskflow.TaskflowProjectGetAPIView.as_view(),
        name='taskflow_project_get',
    ),
    url(
        regex=r'^taskflow/update$',
        view=views_taskflow.TaskflowProjectUpdateAPIView.as_view(),
        name='taskflow_project_update',
    ),
    url(
        regex=r'^taskflow/role/get$',
        view=views_taskflow.TaskflowRoleAssignmentGetAPIView.as_view(),
        name='taskflow_role_get',
    ),
    url(
        regex=r'^taskflow/role/set$',
        view=views_taskflow.TaskflowRoleAssignmentSetAPIView.as_view(),
        name='taskflow_role_set',
    ),
    url(
        regex=r'^taskflow/role/delete$',
        view=views_taskflow.TaskflowRoleAssignmentDeleteAPIView.as_view(),
        name='taskflow_role_delete',
    ),
    url(
        regex=r'^taskflow/settings/get$',
        view=views_taskflow.TaskflowProjectSettingsGetAPIView.as_view(),
        name='taskflow_settings_get',
    ),
    url(
        regex=r'^taskflow/settings/set$',
        view=views_taskflow.TaskflowProjectSettingsSetAPIView.as_view(),
        name='taskflow_settings_set',
    ),
]

urlpatterns = urls_ui + urls_ajax + urls_api + urls_taskflow



"""SODAR Taskflow API for Django apps"""

import logging
import requests
from uuid import UUID

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

# Projectroles dependency
from projectroles.models import RoleAssignment, SODAR_CONSTANTS


logger = logging.getLogger(__name__)

# SODAR constants
PROJECT_TYPE_PROJECT = SODAR_CONSTANTS['PROJECT_TYPE_PROJECT']

# Local constants
HEADERS = {'Content-Type': 'application/json'}


class TaskflowAPI:
    """SODAR Taskflow API to be used by Django apps"""

    class FlowSubmitException(Exception):
        """SODAR Taskflow submission exception"""

        pass

    class CleanupException(Exception):
        """SODAR Taskflow cleanup exception"""

        pass

    def __init__(self):
        self.taskflow_url = '{}:{}'.format(
            getattr(settings, 'TASKFLOW_BACKEND_HOST', ''),
            getattr(settings, 'TASKFLOW_BACKEND_PORT', ''),
        )

    def submit(
        self,
        project_uuid,
        flow_name,
        flow_data,
        request=None,
        targets=None,
        request_mode='sync',
        timeline_uuid=None,
        force_fail=False,
        sodar_url=None,
    ):
        """
        Submit taskflow for SODAR project data modification.

        :param project_uuid: UUID of the project (UUID object or string)
        :param flow_name: Name of flow to be executed (string)
        :param flow_data: Input data for flow execution (dict)
        :param request: Request object (optional)
        :param targets: Names of backends to sync with (list)
        :param request_mode: "sync" or "async"
        :param timeline_uuid: UUID of corresponding timeline event (optional)
        :param force_fail: Make flow fail on purpose (boolean, default False)
        :param sodar_url: URL of SODAR server (optional, for testing)
        :return: Boolean
        :raise: FlowSubmitException if submission fails
        """
        if not targets:
            targets = settings.TASKFLOW_TARGETS
        url = self.taskflow_url + '/submit'

        # Format UUIDs in flow_data
        for k, v in flow_data.items():
            if isinstance(v, UUID):
                flow_data[k] = str(v)

        data = {
            'project_uuid': str(project_uuid),
            'flow_name': flow_name,
            'flow_data': flow_data,
            'request_mode': request_mode,
            'targets': targets,
            'force_fail': force_fail,
            'timeline_uuid': str(timeline_uuid),
            'sodar_secret': settings.TASKFLOW_SODAR_SECRET,
        }

        # Add the "test_mode" parameter
        data['test_mode'] = settings.TASKFLOW_TEST_MODE

        # HACK: Add overriding URL for test server
        if sodar_url:
            data['sodar_url'] = sodar_url
        elif request:
            if request.POST.get('sodar_url'):
                data['sodar_url'] = request.POST['sodar_url']
            elif request.GET.get('sodar_url'):
                data['sodar_url'] = request.GET['sodar_url']
            elif hasattr(request, 'data') and request.data.get('sodar_url'):
                data['sodar_url'] = request.data['sodar_url']

        logger.debug('Submit data: {}'.format(data))
        response = requests.post(url, json=data, headers=HEADERS)

        if response.status_code == 200 and bool(response.text) is True:
            logger.debug('Submit OK')
            return True

        else:
            logger.error('Submit failed: {}'.format(response.text))
            raise self.FlowSubmitException(
                self.get_error_msg(flow_name, response.text)
            )

    def use_taskflow(self, project):
        """
        Check whether taskflow use is allowed with a project.

        :param project: Project object
        :return: Boolean
        """
        return True if project.type == PROJECT_TYPE_PROJECT else False

    def cleanup(self):
        """
        Send a cleanup command to SODAR Taskflow. Only allowed in test mode.

        :return: Boolean
        :raise: ImproperlyConfigured if TASKFLOW_TEST_MODE is not set True
        :raise: CleanupException if SODAR Taskflow raises an error
        """
        if not settings.TASKFLOW_TEST_MODE:
            raise ImproperlyConfigured(
                'TASKFLOW_TEST_MODE not True, cleanup command not allowed'
            )

        url = self.taskflow_url + '/cleanup'
        data = {'test_mode': settings.TASKFLOW_TEST_MODE}

        response = requests.post(url, json=data, headers=HEADERS)

        if response.status_code == 200:
            logger.debug('Cleanup OK')
            return True

        else:
            logger.debug('Cleanup failed: {}'.format(response.text))
            raise self.CleanupException(response.text)

    def get_error_msg(self, flow_name, submit_info):
        """
        Return a printable version of a SODAR Taskflow error message.

        :param flow_name: Name of submitted flow
        :param submit_info: Returned information from SODAR Taskflow
        :return: String
        """
        return 'Taskflow "{}" failed! Reason: "{}"'.format(
            flow_name, submit_info[:256]
        )

    @classmethod
    def get_inherited_roles(cls, project, user, roles=None):
        """
        Return list of inherited owner roles to be used in taskflow sync.

        :param project: Project object
        :param user: User object
        :pram roles: Previously collected roles (optional, list or None)
        :return: List of dicts
        """
        if roles is None:
            roles = []

        # TODO: Remove support for legacy roles in v0.9 (see #506)
        if (
            project.type == PROJECT_TYPE_PROJECT
            and not RoleAssignment.objects.filter(project=project, user=user)
        ):
            r = {
                'project_uuid': str(project.sodar_uuid),
                'username': user.username,
            }

            if r not in roles:  # Avoid unnecessary dupes
                roles.append(r)

        for child in project.get_children():
            roles = cls.get_inherited_roles(child, user, roles)

        return roles

    @classmethod
    def get_inherited_users(cls, project, roles=None):
        """
        Return list of all inherited users within a project and its children, to
        be used in taskflow sync.

        :param project: Project object
        :pram roles: Previously collected roles (optional, list or None)
        :return: List of dicts
        """

        if roles is None:
            roles = []

        if project.type == PROJECT_TYPE_PROJECT:
            i_owners = [a.user for a in project.get_owners(inherited_only=True)]
            all_users = [a.user for a in project.get_all_roles(inherited=False)]

            for u in [u for u in i_owners if u not in all_users]:
                roles.append(
                    {
                        'project_uuid': str(project.sodar_uuid),
                        'username': u.username,
                    }
                )

        for child in project.get_children():
            roles = cls.get_inherited_users(child, roles)

        return roles



"""Permission tests for the filesfolders app"""

from django.test import override_settings
from django.urls import reverse

# Projectroles dependency
from projectroles.app_settings import AppSettingAPI
from projectroles.models import SODAR_CONSTANTS
from projectroles.tests.test_permissions import TestProjectPermissionBase

from filesfolders.tests.test_models import (
    FileMixin,
    FolderMixin,
    HyperLinkMixin,
)

# SODAR constants
PROJECT_ROLE_OWNER = SODAR_CONSTANTS['PROJECT_ROLE_OWNER']
PROJECT_ROLE_DELEGATE = SODAR_CONSTANTS['PROJECT_ROLE_DELEGATE']
PROJECT_ROLE_CONTRIBUTOR = SODAR_CONSTANTS['PROJECT_ROLE_CONTRIBUTOR']
PROJECT_ROLE_GUEST = SODAR_CONSTANTS['PROJECT_ROLE_GUEST']
PROJECT_TYPE_CATEGORY = SODAR_CONSTANTS['PROJECT_TYPE_CATEGORY']
PROJECT_TYPE_PROJECT = SODAR_CONSTANTS['PROJECT_TYPE_PROJECT']

# Local constants
APP_NAME = 'filesfolders'
SECRET = '7dqq83clo2iyhg29hifbor56og6911r5'


# App settings API
app_settings = AppSettingAPI()


class TestListPermissions(TestProjectPermissionBase):
    """Tests for the file list view"""

    def test_list(self):
        """Test file list"""
        url = reverse(
            'filesfolders:list', kwargs={'project': self.project.sodar_uuid}
        )
        good_users = [
            self.superuser,
            self.owner_as.user,
            self.delegate_as.user,
            self.contributor_as.user,
            self.guest_as.user,
        ]
        bad_users = [self.user_no_roles, self.anonymous]
        self.assert_response(url, good_users, 200)
        self.assert_response(url, bad_users, 302)
        # Test public project
        self.project.set_public()
        self.assert_response(url, self.user_no_roles, 200)
        self.assert_response(url, self.anonymous, 302)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_list_anon(self):
        """Test file list with anonynomus access"""
        url = reverse(
            'filesfolders:list', kwargs={'project': self.project.sodar_uuid}
        )
        self.project.set_public()
        self.assert_response(url, self.anonymous, 200)


class TestFolderPermissions(FolderMixin, TestProjectPermissionBase):
    """Tests for Folder views"""

    def setUp(self):
        super().setUp()

        self.folder = self._make_folder(
            name='folder',
            project=self.project,
            folder=None,
            owner=self.owner_as.user,  # Project owner is the owner of folder
            description='',
        )

    def test_folder_create(self):
        """Test folder creation"""
        url = reverse(
            'filesfolders:folder_create',
            kwargs={'project': self.project.sodar_uuid},
        )
        good_users = [
            self.superuser,
            self.owner_as.user,
            self.delegate_as.user,
            self.contributor_as.user,
        ]
        bad_users = [self.guest_as.user, self.user_no_roles, self.anonymous]
        self.assert_response(url, good_users, 200)
        self.assert_response(url, bad_users, 302)
        # Test public project
        self.project.set_public()
        self.assert_response(url, bad_users, 302)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_folder_create_anon(self):
        """Test folder creation with anonymous access"""
        url = reverse(
            'filesfolders:folder_create',
            kwargs={'project': self.project.sodar_uuid},
        )
        self.project.set_public()
        self.assert_response(url, self.anonymous, 302)

    def test_folder_create_category(self):
        """Test folder creation under category"""
        url = reverse(
            'filesfolders:folder_create',
            kwargs={'project': self.category.sodar_uuid},
        )
        bad_users = [
            self.anonymous,
            self.superuser,
            self.owner_as.user,
            self.delegate_as.user,
            self.contributor_as.user,
            self.guest_as.user,
            self.user_no_roles,
        ]
        self.assert_response(url, bad_users, 302)

    def test_folder_update(self):
        """Test folder updating"""
        url = reverse(
            'filesfolders:folder_update',
            kwargs={'item': self.folder.sodar_uuid},
        )
        good_users = [self.superuser, self.owner_as.user, self.delegate_as.user]
        bad_users = [
            self.contributor_as.user,  # NOTE: not the owner of the folder
            self.anonymous,
            self.guest_as.user,
            self.user_no_roles,
        ]
        self.assert_response(url, good_users, 200)
        self.assert_response(url, bad_users, 302)
        # Test public project
        self.project.set_public()
        self.assert_response(url, bad_users, 302)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_folder_update_anon(self):
        """Test folder updating with anonymous access"""
        url = reverse(
            'filesfolders:folder_update',
            kwargs={'item': self.folder.sodar_uuid},
        )
        self.project.set_public()
        self.assert_response(url, self.anonymous, 302)

    def test_folder_delete(self):
        """Test folder deletion"""
        url = reverse(
            'filesfolders:folder_delete',
            kwargs={'item': self.folder.sodar_uuid},
        )
        good_users = [self.superuser, self.owner_as.user, self.delegate_as.user]
        bad_users = [
            self.contributor_as.user,  # NOTE: not the owner of the folder
            self.anonymous,
            self.guest_as.user,
            self.user_no_roles,
        ]
        self.assert_response(url, good_users, 200)
        self.assert_response(url, bad_users, 302)
        # Test public project
        self.project.set_public()
        self.assert_response(url, bad_users, 302)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_folder_delete_anon(self):
        """Test folder deletion with anonymous access"""
        url = reverse(
            'filesfolders:folder_delete',
            kwargs={'item': self.folder.sodar_uuid},
        )
        self.project.set_public()
        self.assert_response(url, self.anonymous, 302)


class TestFilePermissions(FileMixin, TestProjectPermissionBase):
    """Tests for File views"""

    def setUp(self):
        super().setUp()

        app_settings.set_app_setting(
            APP_NAME, 'allow_public_links', True, project=self.project
        )

        self.file_content = bytes('content'.encode('utf-8'))

        # Init file
        self.file = self._make_file(
            name='file.txt',
            file_name='file.txt',
            file_content=self.file_content,
            project=self.project,
            folder=None,
            owner=self.owner_as.user,  # Project owner is the file owner
            description='',
            public_url=True,
            secret=SECRET,
        )

    def test_file_create(self):
        """Test file creation"""
        url = reverse(
            'filesfolders:file_create',
            kwargs={'project': self.project.sodar_uuid},
        )
        good_users = [
            self.superuser,
            self.owner_as.user,
            self.delegate_as.user,
            self.contributor_as.user,
        ]
        bad_users = [self.anonymous, self.guest_as.user, self.user_no_roles]
        self.assert_response(url, good_users, 200)
        self.assert_response(url, bad_users, 302)
        # Test public project
        self.project.set_public()
        self.assert_response(url, bad_users, 302)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_file_create_anon(self):
        """Test file creation with anonymous access"""
        url = reverse(
            'filesfolders:file_create',
            kwargs={'project': self.project.sodar_uuid},
        )
        self.project.set_public()
        self.assert_response(url, self.anonymous, 302)

    def test_file_create_category(self):
        """Test file creation under category"""
        url = reverse(
            'filesfolders:file_create',
            kwargs={'project': self.category.sodar_uuid},
        )
        bad_users = [
            self.anonymous,
            self.superuser,
            self.owner_as.user,
            self.delegate_as.user,
            self.contributor_as.user,
            self.guest_as.user,
            self.user_no_roles,
        ]
        self.assert_response(url, bad_users, 302)

    def test_file_update(self):
        """Test file updating"""
        url = reverse(
            'filesfolders:file_update', kwargs={'item': self.file.sodar_uuid}
        )
        good_users = [self.superuser, self.owner_as.user, self.delegate_as.user]
        bad_users = [
            self.contributor_as.user,  # NOTE: not the owner of the file
            self.anonymous,
            self.guest_as.user,
            self.user_no_roles,
        ]
        self.assert_response(url, good_users, 200)
        self.assert_response(url, bad_users, 302)
        # Test public project
        self.project.set_public()
        self.assert_response(url, bad_users, 302)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_file_update_anon(self):
        """Test file updating with anonymous access"""
        url = reverse(
            'filesfolders:file_update', kwargs={'item': self.file.sodar_uuid}
        )
        self.project.set_public()
        self.assert_response(url, self.anonymous, 302)

    def test_file_delete(self):
        """Test file deletion"""
        url = reverse(
            'filesfolders:file_delete', kwargs={'item': self.file.sodar_uuid}
        )
        good_users = [self.superuser, self.owner_as.user, self.delegate_as.user]
        bad_users = [
            self.contributor_as.user,  # NOTE: not the owner of the file
            self.anonymous,
            self.guest_as.user,
            self.user_no_roles,
        ]
        self.assert_response(url, good_users, 200)
        self.assert_response(url, bad_users, 302)
        # Test public project
        self.project.set_public()
        self.assert_response(url, bad_users, 302)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_file_delete_anon(self):
        """Test file deletion with anonymous access"""
        url = reverse(
            'filesfolders:file_delete', kwargs={'item': self.file.sodar_uuid}
        )
        self.project.set_public()
        self.assert_response(url, self.anonymous, 302)

    def test_file_public_link(self):
        """Test generation of a public URL to a file"""
        url = reverse(
            'filesfolders:file_public_link',
            kwargs={'file': self.file.sodar_uuid},
        )
        good_users = [
            self.superuser,
            self.owner_as.user,
            self.delegate_as.user,
            self.contributor_as.user,
        ]
        bad_users = [self.anonymous, self.guest_as.user, self.user_no_roles]
        self.assert_response(url, good_users, 200)
        self.assert_response(url, bad_users, 302)
        # Test public project
        self.project.set_public()
        self.assert_response(url, bad_users, 302)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_file_public_link_anon(self):
        """Test generation of a public URL to a file with anonymous access"""
        url = reverse(
            'filesfolders:file_public_link',
            kwargs={'file': self.file.sodar_uuid},
        )
        self.project.set_public()
        self.assert_response(url, self.anonymous, 302)

    def test_file_serve(self):
        """Test file serving for authenticated users"""
        url = reverse(
            'filesfolders:file_serve',
            kwargs={'file': self.file.sodar_uuid, 'file_name': self.file.name},
        )
        good_users = [
            self.superuser,
            self.owner_as.user,
            self.delegate_as.user,
            self.contributor_as.user,
            self.guest_as.user,
        ]
        bad_users = [self.anonymous, self.user_no_roles]
        self.assert_response(url, good_users, 200)
        self.assert_response(url, bad_users, 302)
        # Test public project
        self.project.set_public()
        self.assert_response(url, self.user_no_roles, 200)
        self.assert_response(url, self.anonymous, 302)

    def test_file_serve_public(self):
        """Test public file serving"""
        url = reverse(
            'filesfolders:file_serve_public',
            kwargs={'secret': SECRET, 'file_name': self.file.name},
        )
        good_users = [
            self.superuser,
            self.owner_as.user,
            self.delegate_as.user,
            self.contributor_as.user,
            self.guest_as.user,
            self.anonymous,
            self.user_no_roles,
        ]
        self.assert_response(url, good_users, 200)

    def test_file_serve_public_disabled(self):
        """Test public file serving if not allowed in project, should fail"""
        app_settings.set_app_setting(
            APP_NAME, 'allow_public_links', False, project=self.project
        )
        url = reverse(
            'filesfolders:file_serve_public',
            kwargs={'secret': SECRET, 'file_name': self.file.name},
        )
        bad_users = [
            self.superuser,
            self.owner_as.user,
            self.delegate_as.user,
            self.contributor_as.user,
            self.guest_as.user,
            self.user_no_roles,
        ]

        for user in bad_users:
            with self.login(user):
                response = self.client.get(url)
                self.assertEqual(response.status_code, 400)

        # Anonymous
        response = self.client.get(url)
        self.assertEqual(response.status_code, 400)


class TestHyperLinkPermissions(HyperLinkMixin, TestProjectPermissionBase):
    """Tests for HyperLink views"""

    def setUp(self):
        super().setUp()

        # Init link
        self.hyperlink = self._make_hyperlink(
            name='Link',
            url='http://www.google.com/',
            project=self.project,
            folder=None,
            owner=self.user_owner,
            description='',
        )

    def test_hyperlink_create(self):
        """Test hyperlink creation"""
        url = reverse(
            'filesfolders:hyperlink_create',
            kwargs={'project': self.project.sodar_uuid},
        )
        good_users = [
            self.superuser,
            self.owner_as.user,
            self.delegate_as.user,
            self.contributor_as.user,
        ]
        bad_users = [self.anonymous, self.guest_as.user, self.user_no_roles]
        self.assert_response(url, good_users, 200)
        self.assert_response(url, bad_users, 302)
        # Test public project
        self.project.set_public()
        self.assert_response(url, bad_users, 302)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_hyperlink_create_anon(self):
        """Test hyperlink creation with anonymous access"""
        url = reverse(
            'filesfolders:hyperlink_create',
            kwargs={'project': self.project.sodar_uuid},
        )
        self.project.set_public()
        self.assert_response(url, self.anonymous, 302)

    def test_hyperlink_create_category(self):
        """Test hyperlink creation under category"""
        url = reverse(
            'filesfolders:hyperlink_create',
            kwargs={'project': self.category.sodar_uuid},
        )
        bad_users = [
            self.anonymous,
            self.superuser,
            self.owner_as.user,
            self.delegate_as.user,
            self.contributor_as.user,
            self.guest_as.user,
            self.user_no_roles,
        ]
        self.assert_response(url, bad_users, 302)

    def test_hyperlink_update(self):
        """Test hyperlink updating"""
        url = reverse(
            'filesfolders:hyperlink_update',
            kwargs={'item': self.hyperlink.sodar_uuid},
        )
        good_users = [self.superuser, self.owner_as.user, self.delegate_as.user]
        bad_users = [
            self.contributor_as.user,  # NOTE: not the owner of the link
            self.anonymous,
            self.guest_as.user,
            self.user_no_roles,
        ]
        self.assert_response(url, good_users, 200)
        self.assert_response(url, bad_users, 302)
        # Test public project
        self.project.set_public()
        self.assert_response(url, bad_users, 302)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_hyperlink_update_anon(self):
        """Test hyperlink updating with anonymous access"""
        url = reverse(
            'filesfolders:hyperlink_update',
            kwargs={'item': self.hyperlink.sodar_uuid},
        )
        self.project.set_public()
        self.assert_response(url, self.anonymous, 302)

    def test_hyperlink_delete(self):
        """Test hyperlink deletion"""
        url = reverse(
            'filesfolders:hyperlink_delete',
            kwargs={'item': self.hyperlink.sodar_uuid},
        )
        good_users = [self.superuser, self.owner_as.user, self.delegate_as.user]
        bad_users = [
            self.contributor_as.user,  # NOTE: not the owner of the link
            self.anonymous,
            self.guest_as.user,
            self.user_no_roles,
        ]
        self.assert_response(url, good_users, 200)
        self.assert_response(url, bad_users, 302)
        # Test public project
        self.project.set_public()
        self.assert_response(url, bad_users, 302)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_hyperlink_delete_anon(self):
        """Test hyperlink deletion with anonymous access"""
        url = reverse(
            'filesfolders:hyperlink_delete',
            kwargs={'item': self.hyperlink.sodar_uuid},
        )
        self.project.set_public()
        self.assert_response(url, self.anonymous, 302)


class TestBatchPermissions(FolderMixin, TestProjectPermissionBase):
    """Tests for batch editing views"""

    def setUp(self):
        super().setUp()

        self.folder = self._make_folder(
            name='folder',
            project=self.project,
            folder=None,
            owner=self.owner_as.user,  # Project owner is the owner of folder
            description='',
        )

    def test_batch_edit(self):
        """Test access to batch editing confirmation"""
        url = reverse(
            'filesfolders:batch_edit',
            kwargs={'project': self.project.sodar_uuid},
        )

        # NOTE: Contributor is OK as checks for object perms happen after POST
        good_users = [
            self.superuser,
            self.owner_as.user,
            self.delegate_as.user,
            self.contributor_as.user,
        ]
        bad_users = [
            # self.anonymous,
            self.guest_as.user,
            self.user_no_roles,
        ]

        post_data = {
            'batch-action': 'delete',
            'user-confirmed': '0',
            'batch_item_Folder_{}'.format(self.folder.sodar_uuid): '1',
        }

        for user in good_users:
            with self.login(user):
                response = self.client.post(url, post_data)
                self.assertEqual(response.status_code, 200)

        for user in bad_users:
            with self.login(user):
                response = self.client.post(url, post_data)
                self.assertEqual(response.status_code, 302)

        # Test public project
        self.project.set_public()
        for user in bad_users:
            with self.login(user):
                response = self.client.post(url, post_data)
                self.assertEqual(response.status_code, 302)



import sys

from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.management.base import BaseCommand, CommandError

# Projectroles dependency
from projectroles.management.logging import ManagementCommandLogger
from projectroles.models import Project, Role, SODAR_CONSTANTS
from projectroles.plugins import get_active_plugins, get_backend_api


# SODAR constants
PROJECT_TYPE_PROJECT = SODAR_CONSTANTS['PROJECT_TYPE_PROJECT']
PROJECT_TYPE_CATEGORY = SODAR_CONSTANTS['PROJECT_TYPE_CATEGORY']


TARGETS = settings.TASKFLOW_TARGETS
TARGETS.remove('sodar')  # Exclude SODAR from sync as data is already here


logger = ManagementCommandLogger(__name__)
User = get_user_model()


class Command(BaseCommand):
    help = 'Submits missing project data to external storage'

    def add_arguments(self, parser):
        pass

    def _submit_sync(self, app_name, sync_data, raise_exception=False):
        """Submit flows found in an app's sync_data structure"""

        for item in sync_data:
            project = Project.objects.get(sodar_uuid=item['project_uuid'])
            logger.debug(
                'Syncing flow "{}" by {} for "{}" ({})'.format(
                    item['flow_name'],
                    app_name,
                    project.title,
                    project.sodar_uuid,
                )
            )
            try:
                self.taskflow.submit(
                    project_uuid=item['project_uuid'],
                    flow_name=item['flow_name'],
                    flow_data=item['flow_data'],
                    targets=TARGETS,
                )
            except self.taskflow.FlowSubmitException as ex:
                logger.error('Exception raised by flow: {}'.format(ex))
                # If we don't want to continue on failure
                if raise_exception:
                    raise ex

    def _sync_projects(self):
        """Synchronize projects and roles (must be called first!)"""
        logger.info('Synchronizing project data with taskflow...')
        logger.info('Target(s) = ' + ', '.join([t for t in TARGETS]))

        # Only sync PROJECT type projects as we (currently) don't have any
        # use for CATEGORY projects in taskflow
        projects = Project.objects.filter(type=PROJECT_TYPE_PROJECT).order_by(
            'pk'
        )
        project_sync_data = []
        role_sync_data = []

        for project in projects:
            owner_as = project.get_owner()
            if not owner_as:  # This should not happen unless the db is corrupt
                logger.error(
                    'No owner assignment for project "{}" ({})'.format(
                        project.title, project.sodar_uuid
                    )
                )
                continue

            # Create project
            project_sync_data.append(
                {
                    'project_uuid': str(project.sodar_uuid),
                    'project_title': project.title,
                    'flow_name': 'project_create',
                    'flow_data': {
                        'project_title': project.title,
                        'project_description': project.description,
                        'parent_uuid': str(project.parent.sodar_uuid)
                        if project.parent
                        else 0,
                        'owner_username': owner_as.user.username,
                        'owner_uuid': str(owner_as.user.sodar_uuid),
                        'owner_role_pk': owner_as.role.pk,
                    },
                }
            )

            # Set up roles
            role_sync_data.append(
                {
                    'project_uuid': str(project.sodar_uuid),
                    'project_title': project.title,
                    'flow_name': 'role_sync_delete_all',
                    'flow_data': {'owner_username': owner_as.user.username},
                }
            )
            for role_as in project.roles.exclude(
                role=Role.objects.get(
                    name=SODAR_CONSTANTS['PROJECT_ROLE_OWNER']
                )
            ):
                role_sync_data.append(
                    {
                        'project_uuid': str(project.sodar_uuid),
                        'project_title': project.title,
                        'flow_name': 'role_update',
                        'flow_data': {
                            'username': role_as.user.username,
                            'user_uuid': str(role_as.user.sodar_uuid),
                            'role_pk': str(role_as.role.pk),
                        },
                    }
                )

        self._submit_sync(
            'projectroles', project_sync_data, raise_exception=True
        )
        self._submit_sync('projectroles', role_sync_data, raise_exception=False)

    def _sync_inherited_owners(self):
        """Synchronize inherited owner permissions in iRODS"""
        timeline = get_backend_api('timeline_backend')
        roles_add = []
        roles_delete = []

        if not timeline:
            logger.warning(
                'Timeline backend not enabled, unable to sync '
                'inherited owner roles in SODAR Taskflow'
            )
        else:
            logger.info('Retrieving added/removed inherited owners..')
            ProjectEvent, ProjectEventObjectRef = timeline.get_models()

            for category in Project.objects.filter(type=PROJECT_TYPE_CATEGORY):
                # Get roles to add
                roles_add = self.taskflow.get_inherited_roles(
                    category, category.get_owner().user, roles_add
                )

                # Get previous owners to remove
                tl_users = (
                    ProjectEventObjectRef.objects.filter(
                        label='prev_owner',
                        event__in=ProjectEvent.objects.filter(
                            project=category, event_name='role_owner_transfer'
                        ),
                    )
                    .exclude(object_uuid=category.get_owner().user.sodar_uuid)
                    .values_list('name', flat=True)
                    .distinct()
                )

                for u_name in tl_users:
                    user = User.objects.filter(username=u_name).first()
                    if not user:
                        continue
                    roles_delete = self.taskflow.get_inherited_roles(
                        category, user, roles_delete
                    )

        if roles_add or roles_delete:
            logger.info('Changes in inherited owners found, synchronizing..')
            try:
                self.taskflow.submit(
                    project_uuid=None,
                    flow_name='role_update_irods_batch',
                    flow_data={
                        'roles_add': roles_add,
                        'roles_delete': roles_delete,
                    },
                )
            except Exception as ex:
                logger.error(
                    'Error synchronizing inherited owners: {}'.format(ex)
                )
            logger.info('Inherited owner permissions synchronized.')

    def _sync_apps(self):
        """Run taskflow synchronization methods in project app plugins"""
        plugins = get_active_plugins(plugin_type='project_app')
        for plugin in plugins:
            sync_data = plugin.get_taskflow_sync_data()
            logger.info('Synchronizing app "{}"...'.format(plugin.name))
            if sync_data:
                self._submit_sync(plugin.name, sync_data, raise_exception=False)
            else:
                logger.info('Nothing to synchronize.')

    def handle(self, *args, **options):
        """Run management command"""
        if 'taskflow' not in settings.ENABLED_BACKEND_PLUGINS:
            logger.error('Taskflow not enabled in settings, cancelled!')
            raise CommandError

        self.taskflow = get_backend_api('taskflow')
        if not self.taskflow:
            logger.error('Taskflow backend plugin not available, cancelled!')
            raise CommandError

        # Projectroles sync
        # NOTE: For projectroles, this is done here as projects must be created
        #       or we can not continue with sync.. Also, removed projects are
        #       NOT deleted automatically (they shouldn't be deleted anyway).
        #       We first set up the projects and exit if syncing them fails.
        try:
            self._sync_projects()
        except Exception as ex:
            logger.error('Exception in project sync: {}'.format(ex))
            logger.error('Project sync failed! Unable to continue, exiting..')
            sys.exit(1)

        # Inherited Owner Sync
        self._sync_inherited_owners()
        # App sync
        self._sync_apps()
        logger.info('Project data synchronized.')



from django.conf.urls import url

from tokens import views

app_name = 'tokens'

urlpatterns = [
    url(regex=r'^$', view=views.UserTokenListView.as_view(), name='list'),
    url(
        regex=r'^create/',
        view=views.UserTokenCreateView.as_view(),
        name='create',
    ),
    url(
        regex=r'^delete/(?P<pk>.+)$',
        view=views.UserTokenDeleteView.as_view(),
        name='delete',
    ),
]



"""REST API view tests for the filesfolders app"""

import json

from django.urls import reverse
from test_plus.test import APITestCase

# Projectroles dependency
from projectroles.models import SODAR_CONSTANTS
from projectroles.tests.test_views_api import SODARAPIViewTestMixin
from projectroles.views_api import (
    CORE_API_MEDIA_TYPE,
    CORE_API_DEFAULT_VERSION,
    INVALID_PROJECT_TYPE_MSG,
)

from filesfolders.tests.test_views import ZIP_PATH_NO_FILES, TestViewsBaseMixin
from filesfolders.models import Folder, File, HyperLink


# SODAR constants
PROJECT_TYPE_CATEGORY = SODAR_CONSTANTS['PROJECT_TYPE_CATEGORY']

# Local constants
INVALID_UUID = '11111111-1111-1111-1111-111111111111'


class TestFilesfoldersAPIViewsBase(
    TestViewsBaseMixin, SODARAPIViewTestMixin, APITestCase
):
    """Base class for filesfolders API tests"""

    media_type = CORE_API_MEDIA_TYPE
    api_version = CORE_API_DEFAULT_VERSION

    def setUp(self):
        super().setUp()
        # Get knox token for self.user
        self.knox_token = self.get_token(self.user)


class TestFolderListCreateAPIView(TestFilesfoldersAPIViewsBase):
    """Tests for the FolderListCreateAPIView class"""

    def setUp(self):
        super().setUp()
        self.folder_data = {
            'name': 'New Folder',
            'flag': 'IMPORTANT',
            'description': 'Folder\'s description',
        }

    def test_list_superuser(self):
        """Test GET request listing folders"""
        response = self.request_knox(
            reverse(
                'filesfolders:api_folder_list_create',
                kwargs={'project': self.project.sodar_uuid},
            )
        )
        self.assertEqual(response.status_code, 200, msg=response.data)
        expected = [
            {
                'name': self.folder.name,
                'folder': None,
                'owner': self.get_serialized_user(self.folder.owner),
                'project': str(self.folder.project.sodar_uuid),
                'flag': self.folder.flag,
                'description': self.folder.description,
                'date_modified': self.get_drf_datetime(
                    self.folder.date_modified
                ),
                'sodar_uuid': str(self.folder.sodar_uuid),
            }
        ]
        self.assertEqual(json.loads(response.content), expected)

    def test_create_in_root(self):
        """Test creation of new folder in root"""
        response = self.request_knox(
            reverse(
                'filesfolders:api_folder_list_create',
                kwargs={'project': self.project.sodar_uuid},
            ),
            method='POST',
            data=self.folder_data,
        )

        self.assertEqual(response.status_code, 201, msg=response.data)
        new_folder = Folder.objects.filter(
            sodar_uuid=response.data['sodar_uuid']
        ).first()
        self.assertIsNotNone(new_folder)

        expected = {
            **self.folder_data,
            'folder': None,
            'owner': self.get_serialized_user(self.user),
            'project': str(self.project.sodar_uuid),
            'date_modified': self.get_drf_datetime(new_folder.date_modified),
            'sodar_uuid': str(new_folder.sodar_uuid),
        }
        self.assertEqual(json.loads(response.content), expected)

    def test_create_in_folder(self):
        """Test creation of new folder below other"""
        folder_data = {
            **self.folder_data,
            'folder': str(self.folder.sodar_uuid),
        }
        response = self.request_knox(
            reverse(
                'filesfolders:api_folder_list_create',
                kwargs={'project': self.project.sodar_uuid},
            ),
            method='POST',
            data=folder_data,
        )

        self.assertEqual(response.status_code, 201, msg=response.data)
        new_folder = Folder.objects.filter(
            sodar_uuid=response.data['sodar_uuid']
        ).first()
        self.assertIsNotNone(new_folder)

        expected = {
            **folder_data,
            'owner': self.get_serialized_user(self.user),
            'project': str(self.project.sodar_uuid),
            'date_modified': self.get_drf_datetime(new_folder.date_modified),
            'sodar_uuid': str(new_folder.sodar_uuid),
        }
        self.assertEqual(json.loads(response.content), expected)

    def test_create_in_category(self):
        """Test creation of new folder in a category (should fail)"""
        category = self._make_project(
            'TestCategory', PROJECT_TYPE_CATEGORY, None
        )
        self._make_assignment(category, self.user, self.role_owner)
        response = self.request_knox(
            reverse(
                'filesfolders:api_folder_list_create',
                kwargs={'project': category.sodar_uuid},
            ),
            method='POST',
            data=self.folder_data,
        )
        self.assertEqual(response.status_code, 403, msg=response.data)
        self.assertEqual(
            str(response.data['detail']),
            INVALID_PROJECT_TYPE_MSG.format(project_type=PROJECT_TYPE_CATEGORY),
        )


class TestFolderRetrieveUpdateDestroyAPIView(TestFilesfoldersAPIViewsBase):
    """Tests for the FolderRetrieveUpdateDestroyAPIView class"""

    def test_retrieve(self):
        """Test retrieval of Folder model through API"""
        response = self.request_knox(
            reverse(
                'filesfolders:api_folder_retrieve_update_destroy',
                kwargs={'folder': self.folder.sodar_uuid},
            )
        )

        self.assertEqual(response.status_code, 200, msg=response.data)
        expected = {
            'name': self.folder.name,
            'folder': None,
            'owner': self.get_serialized_user(self.folder.owner),
            'project': str(self.folder.project.sodar_uuid),
            'flag': self.folder.flag,
            'description': self.folder.description,
            'date_modified': self.get_drf_datetime(self.folder.date_modified),
            'sodar_uuid': str(self.folder.sodar_uuid),
        }
        self.assertEqual(json.loads(response.content), expected)

    def test_retrieve_not_found(self):
        """Test retrieval of Folder with invalid UUID"""
        response = self.request_knox(
            reverse(
                'filesfolders:api_folder_retrieve_update_destroy',
                kwargs={'folder': INVALID_UUID},
            )
        )
        self.assertEqual(response.status_code, 404)

    def test_update(self):
        """Test update of Folder model through API"""

        folder_data = {
            'name': 'UPDATED Folder',
            'flag': 'FLAG',
            'description': 'UPDATED Description',
        }
        response = self.request_knox(
            reverse(
                'filesfolders:api_folder_retrieve_update_destroy',
                kwargs={'folder': self.folder.sodar_uuid},
            ),
            method='PUT',
            data=folder_data,
        )

        self.assertEqual(response.status_code, 200, msg=response.data)

        self.folder.refresh_from_db()
        expected = {
            **folder_data,
            'folder': None,
            'owner': self.get_serialized_user(self.folder.owner),
            'project': str(self.folder.project.sodar_uuid),
            'date_modified': self.get_drf_datetime(self.folder.date_modified),
            'sodar_uuid': str(self.folder.sodar_uuid),
        }
        self.assertEqual(json.loads(response.content), expected)

    def test_destroy(self):
        """Test destruction of Folder model through API"""
        response = self.request_knox(
            reverse(
                'filesfolders:api_folder_retrieve_update_destroy',
                kwargs={'folder': self.folder.sodar_uuid},
            ),
            method='DELETE',
        )

        self.assertEqual(response.status_code, 204, msg=response.data)
        self.assertIsNone(response.data)

        with self.assertRaises(Folder.DoesNotExist):
            Folder.objects.get(
                project=self.project, sodar_uuid=self.folder.sodar_uuid
            )


class TestFileListCreateAPIView(TestFilesfoldersAPIViewsBase):
    """Tests for the FileListCreateAPIView class"""

    def setUp(self):
        super().setUp()
        self.file_data = {
            'name': 'New File',
            'flag': 'IMPORTANT',
            'description': 'File\'s description',
            'secret': 'foo',
            'public_url': True,
            'file': open(ZIP_PATH_NO_FILES, 'rb'),
        }

    def tearDown(self):
        self.file_data['file'].close()
        super().tearDown()

    def test_list_superuser(self):
        """Test GET request listing files"""
        response = self.request_knox(
            reverse(
                'filesfolders:api_file_list_create',
                kwargs={'project': self.project.sodar_uuid},
            )
        )

        self.assertEqual(response.status_code, 200, msg=response.data)

        expected = [
            {
                'name': self.file.name,
                'folder': None,
                'owner': self.get_serialized_user(self.file.owner),
                'project': str(self.file.project.sodar_uuid),
                'flag': self.file.flag,
                'description': self.file.description,
                'secret': self.file.secret,
                'public_url': self.file.public_url,
                'date_modified': self.get_drf_datetime(self.file.date_modified),
                'sodar_uuid': str(self.file.sodar_uuid),
            }
        ]
        self.assertEqual(json.loads(response.content), expected)

    def test_create_in_root(self):
        """Test creation of new file in root"""
        response = self.request_knox(
            reverse(
                'filesfolders:api_file_list_create',
                kwargs={'project': self.project.sodar_uuid},
            ),
            method='POST',
            format='multipart',
            data=self.file_data,
        )

        self.assertEqual(response.status_code, 201, msg=response.data)
        new_file = File.objects.filter(
            sodar_uuid=response.data['sodar_uuid']
        ).first()
        self.assertIsNotNone(new_file)
        self.assertNotEqual(new_file.file.file.size, 0)

        expected = {
            **self.file_data,
            'folder': None,
            'owner': self.get_serialized_user(self.user),
            'project': str(self.project.sodar_uuid),
            'secret': new_file.secret,
            'public_url': new_file.public_url,
            'date_modified': self.get_drf_datetime(new_file.date_modified),
            'sodar_uuid': str(new_file.sodar_uuid),
        }
        expected.pop('file')
        self.assertEqual(json.loads(response.content), expected)

    def test_create_in_folder(self):
        """Test creation of a file inside a folder"""
        file_data = {**self.file_data, 'folder': str(self.folder.sodar_uuid)}

        response = self.request_knox(
            reverse(
                'filesfolders:api_file_list_create',
                kwargs={'project': self.project.sodar_uuid},
            ),
            method='POST',
            format='multipart',
            data=file_data,
        )

        self.assertEqual(response.status_code, 201, msg=response.data)
        new_file = File.objects.filter(
            sodar_uuid=response.data['sodar_uuid']
        ).first()
        self.assertIsNotNone(new_file)
        self.assertNotEqual(new_file.file.file.size, 0)

        expected = {
            **file_data,
            'owner': self.get_serialized_user(self.user),
            'project': str(self.project.sodar_uuid),
            'public_url': self.file_data['public_url'],
            'secret': new_file.secret,
            'date_modified': self.get_drf_datetime(new_file.date_modified),
            'sodar_uuid': str(new_file.sodar_uuid),
        }
        expected.pop('file')
        self.assertEqual(json.loads(response.content), expected)

    def test_create_in_category(self):
        """Test creation of new file in a category (should fail)"""
        category = self._make_project(
            'TestCategory', PROJECT_TYPE_CATEGORY, None
        )
        self._make_assignment(category, self.user, self.role_owner)
        response = self.request_knox(
            reverse(
                'filesfolders:api_file_list_create',
                kwargs={'project': category.sodar_uuid},
            ),
            method='POST',
            format='multipart',
            data=self.file_data,
        )
        self.assertEqual(response.status_code, 403, msg=response.data)
        self.assertEqual(
            str(response.data['detail']),
            INVALID_PROJECT_TYPE_MSG.format(project_type=PROJECT_TYPE_CATEGORY),
        )


class TestFileRetrieveUpdateDestroyAPIView(TestFilesfoldersAPIViewsBase):
    """Tests for the FileRetrieveUpdateDestroyAPIView class"""

    def setUp(self):
        super().setUp()
        self.file_data = {
            'name': 'UPDATED File',
            'flag': 'FLAG',
            'description': 'UPDATED description',
            'secret': 'foo',
            'public_url': False,
            'file': open(ZIP_PATH_NO_FILES, 'rb'),
        }

    def tearDown(self):
        self.file_data['file'].close()
        super().tearDown()

    def test_retrieve(self):
        """Test retrieval of File model through API"""
        response = self.request_knox(
            reverse(
                'filesfolders:api_file_retrieve_update_destroy',
                kwargs={'file': self.file.sodar_uuid},
            )
        )

        self.assertEqual(response.status_code, 200, msg=response.data)

        expected = {
            'name': self.file.name,
            'folder': None,
            'owner': self.get_serialized_user(self.file.owner),
            'project': str(self.file.project.sodar_uuid),
            'flag': self.file.flag,
            'description': self.file.description,
            'public_url': self.file.public_url,
            'secret': self.file.secret,
            'date_modified': self.get_drf_datetime(self.file.date_modified),
            'sodar_uuid': str(self.file.sodar_uuid),
        }
        self.assertEqual(json.loads(response.content), expected)

    def test_retrieve_not_found(self):
        """Test retrieval of File with invalid UUID"""
        response = self.request_knox(
            reverse(
                'filesfolders:api_file_retrieve_update_destroy',
                kwargs={'file': INVALID_UUID},
            )
        )
        self.assertEqual(response.status_code, 404)

    def test_update(self):
        """Test update of File model through API"""
        response = self.request_knox(
            reverse(
                'filesfolders:api_file_retrieve_update_destroy',
                kwargs={'file': self.file.sodar_uuid},
            ),
            method='PUT',
            format='multipart',
            data=self.file_data,
        )

        self.assertEqual(response.status_code, 200, msg=response.data)
        old_secret = self.file.secret
        self.file.refresh_from_db()
        self.assertEqual(self.file.name, self.file_data['name'])
        self.assertEqual(self.file.flag, self.file_data['flag'])
        self.assertEqual(self.file.description, self.file_data['description'])
        self.assertNotEqual(
            self.file.secret,
            old_secret,
            msg='Secret should change when public_url flag changes',
        )

        expected = {
            **self.file_data,
            'folder': None,
            'owner': self.get_serialized_user(self.file.owner),
            'project': str(self.file.project.sodar_uuid),
            'public_url': self.file_data['public_url'],
            'secret': self.file.secret,
            'date_modified': self.get_drf_datetime(self.file.date_modified),
            'sodar_uuid': str(self.file.sodar_uuid),
        }
        expected.pop('file')
        self.assertEqual(json.loads(response.content), expected)

    def test_destroy(self):
        """Test destruction of File model through API"""
        response = self.request_knox(
            reverse(
                'filesfolders:api_file_retrieve_update_destroy',
                kwargs={'file': self.file.sodar_uuid},
            ),
            method='DELETE',
        )

        self.assertEqual(response.status_code, 204, msg=response.data)
        self.assertIsNone(response.data)

        with self.assertRaises(File.DoesNotExist):
            File.objects.get(
                project=self.project, sodar_uuid=self.file.sodar_uuid
            )


class TestFileServeAPIView(TestFilesfoldersAPIViewsBase):
    """Tests for the FileServeAPIView class"""

    def test_get(self):
        """Test download of file content"""
        response = self.request_knox(
            reverse(
                'filesfolders:api_file_serve',
                kwargs={'file': self.file.sodar_uuid},
            )
        )
        expected = b'content'
        self.assertEqual(response.status_code, 200, msg=response.content)
        self.assertEqual(response.content, expected)

    def test_get_not_found(self):
        """Test download with invalid UUID"""
        response = self.request_knox(
            reverse(
                'filesfolders:api_file_serve',
                kwargs={'file': INVALID_UUID},
            )
        )
        self.assertEqual(response.status_code, 404)


class TestHyperLinkListCreateAPIView(TestFilesfoldersAPIViewsBase):
    """Tests for the HyperLinkListCreateAPIView class"""

    def setUp(self):
        super().setUp()
        self.hyperlink_data = {
            'name': 'New HyperLink',
            'flag': 'IMPORTANT',
            'description': 'HyperLink\'s description',
            'url': 'http://www.cubi.bihealth.org',
        }

    def test_list_superuser(self):
        """Test GET request listing hyperlinks"""
        response = self.request_knox(
            reverse(
                'filesfolders:api_hyperlink_list_create',
                kwargs={'project': self.project.sodar_uuid},
            )
        )

        self.assertEqual(response.status_code, 200, msg=response.data)

        expected = [
            {
                'name': self.hyperlink.name,
                'folder': None,
                'owner': self.get_serialized_user(self.hyperlink.owner),
                'project': str(self.hyperlink.project.sodar_uuid),
                'flag': self.hyperlink.flag,
                'description': self.hyperlink.description,
                'url': self.hyperlink.url,
                'date_modified': self.get_drf_datetime(
                    self.hyperlink.date_modified
                ),
                'sodar_uuid': str(self.hyperlink.sodar_uuid),
            }
        ]
        self.assertEqual(json.loads(response.content), expected)

    def test_create_in_root(self):
        """Test creation of new hyperlink in root"""
        response = self.request_knox(
            reverse(
                'filesfolders:api_hyperlink_list_create',
                kwargs={'project': self.project.sodar_uuid},
            ),
            method='POST',
            data=self.hyperlink_data,
        )

        self.assertEqual(response.status_code, 201, msg=response.data)
        new_link = HyperLink.objects.filter(
            sodar_uuid=response.data['sodar_uuid']
        ).first()
        self.assertIsNotNone(new_link)

        expected = {
            **self.hyperlink_data,
            'folder': None,
            'owner': self.get_serialized_user(self.user),
            'project': str(self.project.sodar_uuid),
            'date_modified': self.get_drf_datetime(new_link.date_modified),
            'sodar_uuid': str(new_link.sodar_uuid),
        }
        self.assertEqual(json.loads(response.content), expected)

    def test_create_in_folder(self):
        """Test creation of new hyperlink below another"""
        hyperlink_data = {
            **self.hyperlink_data,
            'folder': str(self.folder.sodar_uuid),
        }

        response = self.request_knox(
            reverse(
                'filesfolders:api_hyperlink_list_create',
                kwargs={'project': self.project.sodar_uuid},
            ),
            method='POST',
            data=hyperlink_data,
        )

        self.assertEqual(response.status_code, 201, msg=response.data)
        new_link = HyperLink.objects.filter(
            sodar_uuid=response.data['sodar_uuid']
        ).first()
        self.assertIsNotNone(new_link)

        expected = {
            **hyperlink_data,
            'owner': self.get_serialized_user(self.user),
            'project': str(self.project.sodar_uuid),
            'date_modified': self.get_drf_datetime(new_link.date_modified),
            'sodar_uuid': str(new_link.sodar_uuid),
        }
        self.assertEqual(json.loads(response.content), expected)

    def test_create_in_category(self):
        """Test creation of new hyperlink in a category (should fail)"""
        category = self._make_project(
            'TestCategory', PROJECT_TYPE_CATEGORY, None
        )
        self._make_assignment(category, self.user, self.role_owner)
        response = self.request_knox(
            reverse(
                'filesfolders:api_hyperlink_list_create',
                kwargs={'project': category.sodar_uuid},
            ),
            method='POST',
            data=self.hyperlink_data,
        )
        self.assertEqual(response.status_code, 403, msg=response.data)
        self.assertEqual(
            str(response.data['detail']),
            INVALID_PROJECT_TYPE_MSG.format(project_type=PROJECT_TYPE_CATEGORY),
        )


class TestHyperLinkRetrieveUpdateDestroyAPIView(TestFilesfoldersAPIViewsBase):
    """Tests for the HyperLinkRetrieveUpdateDestroyAPIView class"""

    def test_retrieve(self):
        """Test retrieval of HyperLink model through API"""

        response = self.request_knox(
            reverse(
                'filesfolders:api_hyperlink_retrieve_update_destroy',
                kwargs={'hyperlink': self.hyperlink.sodar_uuid},
            )
        )

        self.assertEqual(response.status_code, 200, msg=response.data)

        expected = {
            'name': self.hyperlink.name,
            'folder': None,
            'owner': self.get_serialized_user(self.hyperlink.owner),
            'project': str(self.hyperlink.project.sodar_uuid),
            'flag': self.hyperlink.flag,
            'description': self.hyperlink.description,
            'url': self.hyperlink.url,
            'date_modified': self.get_drf_datetime(
                self.hyperlink.date_modified
            ),
            'sodar_uuid': str(self.hyperlink.sodar_uuid),
        }
        self.assertEqual(json.loads(response.content), expected)

    def test_retrieve_not_found(self):
        """Test retrieval of HyperLink with invalid UUID"""
        response = self.request_knox(
            reverse(
                'filesfolders:api_hyperlink_retrieve_update_destroy',
                kwargs={'hyperlink': INVALID_UUID},
            )
        )
        self.assertEqual(response.status_code, 404)

    def test_update(self):
        """Test update of HyperLink model through API"""
        hyperlink_data = {
            'name': 'UPDATED HyperLink',
            'flag': 'FLAG',
            'description': 'UPDATED Description',
            'url': 'http://www.bihealth.org',
        }

        response = self.request_knox(
            reverse(
                'filesfolders:api_hyperlink_retrieve_update_destroy',
                kwargs={'hyperlink': self.hyperlink.sodar_uuid},
            ),
            method='PUT',
            data=hyperlink_data,
        )

        self.assertEqual(response.status_code, 200, msg=response.data)
        self.hyperlink.refresh_from_db()
        self.assertEqual(self.hyperlink.name, hyperlink_data['name'])
        self.assertEqual(self.hyperlink.flag, hyperlink_data['flag'])
        self.assertEqual(
            self.hyperlink.description, hyperlink_data['description']
        )

        expected = {
            **hyperlink_data,
            'folder': None,
            'owner': self.get_serialized_user(self.hyperlink.owner),
            'project': str(self.hyperlink.project.sodar_uuid),
            'date_modified': self.get_drf_datetime(
                self.hyperlink.date_modified
            ),
            'sodar_uuid': str(self.hyperlink.sodar_uuid),
        }
        self.assertEqual(json.loads(response.content), expected)

    def test_destroy(self):
        """Test destruction of HyperLink model through API"""
        response = self.request_knox(
            reverse(
                'filesfolders:api_hyperlink_retrieve_update_destroy',
                kwargs={'hyperlink': self.hyperlink.sodar_uuid},
            ),
            method='DELETE',
        )

        self.assertEqual(response.status_code, 204, msg=response.data)
        self.assertIsNone(response.data)



"""Taskflow API views for the timeline app"""

from rest_framework.response import Response

# Projectroles dependency
from projectroles.views_taskflow import BaseTaskflowAPIView

from timeline.models import ProjectEvent


class TaskflowEventStatusSetAPIView(BaseTaskflowAPIView):
    def post(self, request):
        try:
            tl_event = ProjectEvent.objects.get(
                sodar_uuid=request.data['event_uuid']
            )

        except ProjectEvent.DoesNotExist:
            return Response('Timeline event not found', status=404)

        try:
            tl_event.set_status(
                status_type=request.data['status_type'],
                status_desc=request.data['status_desc'],
                extra_data=request.data['extra_data']
                if 'extra_data' in request.data
                else None,
            )

        except TypeError:
            return Response('Invalid status type', status=400)

        return Response('ok', status=200)



"""Plugin point definitions and plugin API for apps based on projectroles"""

from django.conf import settings
from djangoplugins.point import PluginPoint


# Local costants
PLUGIN_TYPES = {
    'project_app': 'ProjectAppPluginPoint',
    'backend': 'BackendPluginPoint',
    'site_app': 'SiteAppPluginPoint',
}

# From djangoplugins
ENABLED = 0
DISABLED = 1
REMOVED = 2


# Plugin points ----------------------------------------------------------------


class ProjectAppPluginPoint(PluginPoint):
    """Projectroles plugin point for registering project specific apps"""

    #: App URLs (will be included in settings by djangoplugins)
    urls = []

    #: App settings definition
    #:
    #: Example ::
    #:
    #:     app_settings = {
    #:         'example_setting': {
    #:             'scope': 'PROJECT',  # PROJECT/USER
    #:             'type': 'STRING',  # STRING/INTEGER/BOOLEAN
    #:             'default': 'example',
    #:             'label': 'Project setting',  # Optional, defaults to name/key
    #:             'placeholder': 'Enter example setting here',  # Optional
    #:             'description': 'Example project setting',  # Optional
    #:             'options': ['example', 'example2'],  # Optional, only for
    #:             settings of type STRING or INTEGER
    #:             'user_modifiable': True,  # Optional, show/hide in forms
    #:             'local': False,  # Optional, show/hide in forms on target
    #:             site, sync value from source to target site if false
    #:         }
    #:     }
    # TODO: Define project specific settings in your app plugin, example above
    app_settings = {}

    # DEPRECATED, will be removed in the next SODAR Core release
    project_settings = {}

    #: Iconify icon
    # TODO: Implement this in your app plugin
    icon = 'mdi:help-rhombus-outline'

    #: Entry point URL ID (must take project sodar_uuid as "project" argument)
    # TODO: Implement this in your app plugin
    entry_point_url_id = 'home'

    #: Description string
    # TODO: Implement this in your app plugin
    description = 'TODO: Write a description for your plugin'

    #: Required permission for accessing the app
    # TODO: Implement this in your app plugin (can be None)
    app_permission = None

    #: Enable or disable general search from project title bar
    # TODO: Implement this in your app plugin
    search_enable = False

    #: List of search object types for the app
    # TODO: Implement this in your app plugin
    search_types = []

    #: Search results template
    # TODO: Implement this in your app plugin
    search_template = None

    #: App card template for the project details page
    # TODO: Implement this in your app plugin
    details_template = None

    #: App card title for the project details page
    # TODO: Implement this in your app plugin (can be None)
    details_title = None

    #: Position in plugin ordering
    # TODO: Implement this in your app plugin (must be an integer)
    plugin_ordering = 50

    #: Optional custom project list column definition
    #:
    #: Example ::
    #:
    #:     project_list_columns = {
    #:         'column_id': {
    #:             'title': 'Column Title',
    #:             'width': 100,  # Desired width of column in pixels
    #:             'description': 'Description',  # Optional description string
    #:             'active': True,  # Boolean for whether the column is active
    #:             'ordering': 50,  # Integer for ordering the columns
    #:             'align': 'left'  # Alignment of content
    #:         }
    #:     }
    # TODO: Define project list column data in your app plugin (optional)
    project_list_columns = {}

    #: Display application for categories in addition to projects
    # TODO: Override this in your app plugin if needed
    category_enable = False

    #: Names of plugin specific Django settings to display in siteinfo
    # TODO: Override this in your app plugin if needed
    info_settings = []

    # NOTE: For projectroles, this is implemented directly in synctaskflow
    def get_taskflow_sync_data(self):
        """
        Return data for synchronizing taskflow operations.

        :return: List of dicts or None.
        """
        '''
        Example of valid return data:
        [
            {
                'flow_name': ''
                'project_pk: ''
                'flow_data': {}
            }
        ]
        '''
        # TODO: Implement this in your app plugin
        return None

    def get_object(self, model, uuid):
        """
        Return object based on a model class and the object's SODAR UUID.

        :param model: Object model class
        :param uuid: sodar_uuid of the referred object
        :return: Model object or None if not found
        :raise: NameError if model is not found
        """
        try:
            return model.objects.get(sodar_uuid=uuid)
        except model.DoesNotExist:
            return None

    def get_object_link(self, model_str, uuid):
        """
        Return URL referring to an object used by the app, along with a label to
        be shown to the user for linking.

        :param model_str: Object class (string)
        :param uuid: sodar_uuid of the referred object
        :return: Dict or None if not found
        """
        obj = self.get_object(eval(model_str), uuid)
        if not obj:
            return None
        # TODO: Implement this in your app plugin
        return None

    def get_extra_data_link(self, _extra_data, _name):
        """Return a link for timeline label starting with 'extra-'"""
        # TODO: Implement this in your app plugin
        return None

    def search(self, search_terms, user, search_type=None, keywords=None):
        """
        Return app items based on one or more search terms, user, optional type
        and optional keywords.

        :param search_terms: Search terms to be joined with the OR operator
                             (list of strings)
        :param user: User object for user initiating the search
        :param search_type: String
        :param keywords: List (optional)
        :return: Dict
        """
        # TODO: Implement this in your app plugin
        # TODO: Implement display of results in the app's search template
        return {
            'all': {  # You can add 1-N lists of result items
                'title': 'Title to be displayed',
                'search_types': [],
                'items': [],
            }
        }

    def update_cache(self, name=None, project=None, user=None):
        """
        Update cached data for this app, limitable to item ID and/or project.

        :param name: Item name to limit update to (string, optional)
        :param project: Project object to limit update to (optional)
        :param user: User object to denote user triggering the update (optional)
        """
        # TODO: Implement this in your app plugin
        return None

    def get_statistics(self):
        """
        Return app statistics as a dict. Should take the form of
        {id: {label, value, url (optional), description (optional)}}.

        :return: Dict
        """
        # TODO: Implement this in your app plugin
        return {}

    def get_project_list_value(self, column_id, project, user):
        """
        Return a value for the optional additional project list column specific
        to a project.

        :param column_id: ID of the column (string)
        :param project: Project object
        :param user: User object (current user)
        :return: String (may contain HTML), integer or None
        """
        # TODO: Implement this in your app plugin (optional)
        return None

    def handle_project_update(self, project, old_data):
        """
        Perform actions to handle project update.
        # NOTE: This is a WIP feature to be altered/expanded in a later release

        :param project: Current project (Project)
        :param old_data: Old project data prior to update (dict)
        """
        # TODO: Implement this in your app plugin (optional)
        pass


class BackendPluginPoint(PluginPoint):
    """Projectroles plugin point for registering backend apps"""

    #: Iconify icon
    # TODO: Implement this in your backend plugin
    icon = 'mdi:help-rhombus-outline'

    #: Description string
    # TODO: Implement this in your backend plugin
    description = 'TODO: Write a description for your plugin'

    #: URL of optional javascript file to be included
    # TODO: Implement this in your backend plugin if applicable
    javascript_url = None

    #: URL of optional css file to be included
    # TODO: Implement this in your backend plugin if applicable
    css_url = None

    #: Names of plugin specific Django settings to display in siteinfo
    # TODO: Override this in your app plugin if needed
    info_settings = []

    def get_api(self):
        """Return API entry point object."""
        # TODO: Implement this in your backend plugin
        raise NotImplementedError

    def get_statistics(self):
        """
        Return backend statistics as a dict. Should take the form of
        {id: {label, value, url (optional), description (optional)}}.

        :return: Dict
        """
        # TODO: Implement this in your backend plugin
        return {}

    def get_object(self, model, uuid):
        """
        Return object based on a model class and the object's SODAR UUID.

        :param model: Object model class
        :param uuid: sodar_uuid of the referred object
        :return: Model object or None if not found
        :raise: NameError if model is not found
        """
        try:
            return model.objects.get(sodar_uuid=uuid)
        except model.DoesNotExist:
            return None

    def get_object_link(self, model_str, uuid):
        """
        Return URL referring to an object used by the app, along with a label to
        be shown to the user for linking.

        :param model_str: Object class (string)
        :param uuid: sodar_uuid of the referred object
        :return: Dict or None if not found
        """
        obj = self.get_object(eval(model_str), uuid)
        if not obj:
            return None
        # TODO: Implement this in your app plugin
        return None

    def get_extra_data_link(self, _extra_data, _name):
        """Return a link for timeline label starting with 'extra-'"""
        # TODO: Implement this in your app plugin
        return None


class SiteAppPluginPoint(PluginPoint):
    """Projectroles plugin point for registering site-wide apps"""

    #: Iconify icon
    # TODO: Implement this in your site app plugin
    icon = 'mdi:help-rhombus-outline'

    #: Description string
    # TODO: Implement this in your site app plugin
    description = 'TODO: Write a description for your plugin'

    #: Entry point URL ID
    # TODO: Implement this in your app plugin
    entry_point_url_id = 'home'

    #: Required permission for displaying the app
    # TODO: Implement this in your site app plugin (can be None)
    app_permission = None

    #: User settings definition
    #:
    #: Example ::
    #:
    #:     app_settings = {
    #:         'example_setting': {
    #:             'scope' : 'USER',  # always USER
    #:             'type': 'STRING',  # STRING/INTEGER/BOOLEAN
    #:             'default': 'example',
    #:             'placeholder': 'Enter example setting here',  # Optional
    #:             'description': 'Example user setting',  # Optional
    #:             'options': ['example', 'example2'],  # Optional, only for
    #:             settings of type STRING or INTEGER
    #:             'user_modifiable': True,  # Optional, show/hide in forms
    #:             'local': False,  # Optional, show/hide in forms on target
    #:             site
    #:         }
    #:     }
    # TODO: Define user specific settings in your app plugin, example above
    app_settings = {}

    #: List of names for plugin specific Django settings to display in siteinfo
    # TODO: Override this in your app plugin if needed
    info_settings = []

    def get_statistics(self):
        """
        Return app statistics as a dict. Should take the form of
        {id: {label, value, url (optional), description (optional)}}.

        :return: Dict
        """
        # TODO: Implement this in your app plugin
        return {}

    def get_messages(self, user=None):
        """
        Return a list of messages to be shown to users.

        :param user: User object (optional)
        :return: List of dicts or and empty list if no messages
        """
        '''
        Example of valid return data:
        return [{
            'content': 'Message content in here, can contain html',
            'color': 'info',        # Corresponds to bg-* in Bootstrap
            'dismissable': True     # False for non-dismissable
        }]
        '''
        # TODO: Implement this in your site app plugin
        return []

    def get_object(self, model, uuid):
        """
        Return object based on a model class and the object's SODAR UUID.

        :param model: Object model class
        :param uuid: sodar_uuid of the referred object
        :return: Model object or None if not found
        :raise: NameError if model is not found
        """
        try:
            return model.objects.get(sodar_uuid=uuid)
        except model.DoesNotExist:
            return None

    def get_object_link(self, model_str, uuid):
        """
        Return URL referring to an object used by the app, along with a label to
        be shown to the user for linking.

        :param model_str: Object class (string)
        :param uuid: sodar_uuid of the referred object
        :return: Dict or None if not found
        """
        obj = self.get_object(eval(model_str), uuid)
        if not obj:
            return None
        # TODO: Implement this in your app plugin
        return None

    def get_extra_data_link(self, _extra_data, _name):
        """Return a link for timeline label starting with 'extra-'"""
        # TODO: Implement this in your app plugin
        return None


# Plugin API -------------------------------------------------------------------


def get_active_plugins(plugin_type='project_app', custom_order=False):
    """
    Return active plugins of a specific type.

    :param plugin_type: "project_app", "site_app" or "backend" (string)
    :param custom_order: Order by plugin_ordering for project apps (boolean)
    :return: List or None
    :raise: ValueError if plugin_type is not recognized
    """
    if plugin_type not in PLUGIN_TYPES.keys():
        raise ValueError(
            'Invalid value for plugin_type. Accepted values: {}'.format(
                ', '.join(PLUGIN_TYPES.keys())
            )
        )

    plugins = eval(PLUGIN_TYPES[plugin_type]).get_plugins()

    if plugins:
        return sorted(
            [
                p
                for p in plugins
                if (
                    p.is_active()
                    and (
                        plugin_type in ['project_app', 'site_app']
                        or p.name in settings.ENABLED_BACKEND_PLUGINS
                    )
                )
            ],
            key=lambda x: x.plugin_ordering
            if custom_order and plugin_type == 'project_app'
            else x.name,
        )

    return None


def change_plugin_status(name, status, plugin_type='app'):
    """
    Change the status of a selected plugin in the database.

    :param name: Plugin name (string)
    :param status: Status (int, see djangoplugins)
    :param plugin_type: Type of plugin ("app", "backend" or "site")
    :raise: ValueError if plugin_type is invalid or plugin with name not found
    """
    # NOTE: Used to forge plugin to a specific status for e.g. testing
    if plugin_type == 'app':
        plugin = ProjectAppPluginPoint.get_plugin(name)
    elif plugin_type == 'backend':
        plugin = BackendPluginPoint.get_plugin(name)
    elif plugin_type == 'site':
        plugin = SiteAppPluginPoint.get_plugin(name)
    else:
        raise ValueError('Invalid plugin_type: "{}"'.format(plugin_type))

    if not plugin:
        raise ValueError(
            'Plugin of type "{}" not found with name "{}"'.format(
                plugin_type, name
            )
        )

    plugin = plugin.get_model()
    plugin.status = status
    plugin.save()


def get_app_plugin(plugin_name, plugin_type=None):
    """
    Return active app plugin.

    :param plugin_name: Plugin name (string)
    :param plugin_type: Plugin type (string or None for all types)
    :return: Plugin object or None if not found
    """
    if plugin_type:
        plugin_types = [PLUGIN_TYPES[plugin_type]]
    else:
        plugin_types = PLUGIN_TYPES.values()
    for t in plugin_types:
        try:
            return eval(t).get_plugin(plugin_name)
        except Exception:
            pass


def get_backend_api(plugin_name, force=False, **kwargs):
    """
    Return backend API object.
    NOTE: May raise an exception from plugin.get_api().

    :param plugin_name: Plugin name (string)
    :param force: Return plugin regardless of status in ENABLED_BACKEND_PLUGINS
    :param kwargs: Optional kwargs for API
    :return: Plugin object or None if not found
    """
    if plugin_name in settings.ENABLED_BACKEND_PLUGINS or force:
        try:
            plugin = BackendPluginPoint.get_plugin(plugin_name)
        except BackendPluginPoint.DoesNotExist:
            return None
        return plugin.get_api(**kwargs) if plugin.is_active() else None


# Plugins within projectroles --------------------------------------------------


class RemoteSiteAppPlugin(SiteAppPluginPoint):
    """Site plugin for remote site and project management"""

    #: Name (slug-safe, used in URLs)
    name = 'remotesites'

    #: Title (used in templates)
    title = 'Remote Site Access'

    #: App URLs (will be included in settings by djangoplugins)
    urls = []

    #: Iconify icon
    icon = 'mdi:cloud-sync'

    #: Description string
    description = 'Management of remote SODAR sites and remote project access'

    #: Entry point URL ID
    entry_point_url_id = 'projectroles:remote_sites'

    #: Required permission for displaying the app
    app_permission = 'userprofile.update_remote'



"""Taskflow view tests for the timeline app"""

import uuid

from django.conf import settings
from django.test import override_settings
from django.urls import reverse

from timeline.tests.test_views import TestViewsBase


class TestTaskflowSetStatusAPIView(TestViewsBase):
    """Tests for the taskflow status setting API view"""

    def setUp(self):
        super().setUp()

        # Init default event
        self.event_init = self.timeline.add_event(
            project=self.project,
            app_name='projectroles',
            user=self.user,
            event_name='test_event',
            description='description',
            extra_data={'test_key': 'test_val'},
            status_type='INIT',
        )

    @override_settings(ENABLED_BACKEND_PLUGINS=['taskflow'])
    def test_set_status(self):
        """Test setting the status of the event"""
        values = {
            'event_uuid': self.event_init.sodar_uuid,
            'status_type': 'OK',
            'status_desc': '',
            'sodar_secret': settings.TASKFLOW_SODAR_SECRET,
        }

        response = self.client.post(
            reverse('timeline:taskflow_status_set'), values
        )

        self.assertEqual(response.status_code, 200)

    @override_settings(ENABLED_BACKEND_PLUGINS=['taskflow'])
    def test_set_invalid_event(self):
        """Test setting the status of the event with an invalid event pk"""
        values = {
            'event_uuid': uuid.uuid4(),
            'status_type': 'OK',
            'status_desc': '',
            'sodar_secret': settings.TASKFLOW_SODAR_SECRET,
        }

        response = self.client.post(
            reverse('timeline:taskflow_status_set'), values
        )

        self.assertEqual(response.status_code, 404)

    @override_settings(ENABLED_BACKEND_PLUGINS=['taskflow'])
    def test_set_invalid_status(self):
        """Test setting the status of the event with an invalid status type"""
        values = {
            'event_uuid': self.event_init.sodar_uuid,
            'status_type': 'ahL4VeerAeth4ohh',
            'status_desc': '',
            'sodar_secret': settings.TASKFLOW_SODAR_SECRET,
        }

        response = self.client.post(
            reverse('timeline:taskflow_status_set'), values
        )

        self.assertEqual(response.status_code, 400)



"""Forms for the projectroles app"""

import json
import logging

from django import forms
from django.conf import settings
from django.contrib import auth
from django.core.exceptions import ValidationError
from django.urls import reverse
from django.utils import timezone

from pagedown.widgets import PagedownWidget
from dal import autocomplete, forward as dal_forward

from projectroles.models import (
    Project,
    Role,
    RoleAssignment,
    ProjectInvite,
    RemoteSite,
    SODAR_CONSTANTS,
    APP_SETTING_VAL_MAXLENGTH,
    CAT_DELIMITER,
)

from projectroles.plugins import get_active_plugins
from projectroles.utils import (
    get_display_name,
    get_user_display_name,
    build_secret,
)
from projectroles.app_settings import AppSettingAPI, APP_SETTING_LOCAL_DEFAULT


User = auth.get_user_model()


# SODAR constants
PROJECT_ROLE_OWNER = SODAR_CONSTANTS['PROJECT_ROLE_OWNER']
PROJECT_ROLE_CONTRIBUTOR = SODAR_CONSTANTS['PROJECT_ROLE_CONTRIBUTOR']
PROJECT_ROLE_DELEGATE = SODAR_CONSTANTS['PROJECT_ROLE_DELEGATE']
PROJECT_ROLE_GUEST = SODAR_CONSTANTS['PROJECT_ROLE_GUEST']
PROJECT_TYPE_CATEGORY = SODAR_CONSTANTS['PROJECT_TYPE_CATEGORY']
PROJECT_TYPE_PROJECT = SODAR_CONSTANTS['PROJECT_TYPE_PROJECT']
SUBMIT_STATUS_OK = SODAR_CONSTANTS['SUBMIT_STATUS_OK']
SUBMIT_STATUS_PENDING = SODAR_CONSTANTS['SUBMIT_STATUS_PENDING']
SUBMIT_STATUS_PENDING_TASKFLOW = SODAR_CONSTANTS['SUBMIT_STATUS_PENDING']
SITE_MODE_SOURCE = SODAR_CONSTANTS['SITE_MODE_SOURCE']
SITE_MODE_TARGET = SODAR_CONSTANTS['SITE_MODE_TARGET']
APP_SETTING_SCOPE_PROJECT = SODAR_CONSTANTS['APP_SETTING_SCOPE_PROJECT']

# Local constants
APP_NAME = 'projectroles'
EMPTY_CHOICE_LABEL = '--------'
PROJECT_TYPE_CHOICES = [
    (None, EMPTY_CHOICE_LABEL),
    (
        PROJECT_TYPE_CATEGORY,
        get_display_name(PROJECT_TYPE_CATEGORY, title=True),
    ),
    (PROJECT_TYPE_PROJECT, get_display_name(PROJECT_TYPE_PROJECT, title=True)),
]


# Base Classes and Mixins ------------------------------------------------------


class SODARFormMixin:
    """General mixin for SODAR form setup and helpers"""

    def __init__(self, *args, **kwargs):
        self.logger = logging.getLogger(__name__)
        super().__init__(*args, **kwargs)

    def add_error(self, field, error):
        """
        Add error to form along with error logging.

        :param field: Name of field (string)
        :param error: Error message(s) (list)
        """
        if isinstance(error, ValidationError):
            log_err = ';'.join(error.messages)
        else:
            log_err = error
        log_msg = 'Field "{}": {}'.format(field, log_err)
        if hasattr(self, 'current_user') and self.current_user:
            log_msg += ' (user={})'.format(self.current_user.username)
        self.logger.error(log_msg)
        super().add_error(field, error)  # Call the error method in Django forms


class SODARForm(SODARFormMixin, forms.Form):
    """Override of Django base form with SODAR Core specific helpers."""


class SODARModelForm(SODARFormMixin, forms.ModelForm):
    """Override of Django model form with SODAR Core specific helpers."""


class SODARPagedownWidget(PagedownWidget):
    class Media:
        css = {'all': ['projectroles/css/pagedown.css']}


# User autocompletion ----------------------------------------------------------


class SODARUserAutocompleteWidget(autocomplete.ModelSelect2):
    """
    Custom Select widget for user field autocompletion which uses the SODAR
    User class sodar_uuid instead of pk.
    """

    # Override function to use sodar_uuid instead of pk
    def filter_choices_to_render(self, selected_choices):
        """Filter out un-selected choices if choices is a QuerySet."""
        self.choices.queryset = self.choices.queryset.filter(
            sodar_uuid__in=[c for c in selected_choices if c]
        )


class SODARUserRedirectWidget(SODARUserAutocompleteWidget):
    """Version of SODARUserAutocompleteWidget to allow redirect on selection"""

    autocomplete_function = 'autocomplete_redirect'


# TODO: Refactor into widget?
def get_user_widget(
    scope='all',
    project=None,
    exclude=None,
    forward=None,
    url=None,
    widget_class=None,
):
    """
    Get an user autocomplete widget for your form.

    :param scope: Scope of users to include: "all"/"project"/"project_exclude"
    :param project: Project object or project UUID string (optional)
    :param exclude: List of User objects or User UUIDs to exclude
    :param forward: Parameters to forward to autocomplete view (optional)
    :param url: Autocomplete ajax class override (optional)
    :param widget_class: Widget class override (optional)
    :return: SODARUserAutocompleteWidget or an overridden widget class
    """
    url = url or 'projectroles:ajax_autocomplete_user'
    wg = {'url': url, 'forward': [dal_forward.Const(scope, 'scope')]}

    if project:
        p_uuid = (
            str(project.sodar_uuid)
            if isinstance(project, Project)
            else str(project)
        )
        wg['forward'].append(dal_forward.Const(p_uuid, 'project'))
    if forward and isinstance(forward, list):
        wg['forward'] += forward
    if exclude:
        wg['forward'].append(
            dal_forward.Const(
                [
                    str(u.sodar_uuid) if isinstance(u, User) else u
                    for u in exclude
                ],
                'exclude',
            )
        )
    if widget_class:
        return widget_class(**wg)

    return SODARUserAutocompleteWidget(**wg)


class SODARUserChoiceField(forms.ModelChoiceField):
    """User choice field to be used with SODAR User objects and autocomplete"""

    def __init__(
        self,
        scope='all',
        project=None,
        exclude=None,
        forward=None,
        url=None,
        widget_class=None,
        *args,
        **kwargs
    ):
        """
        Override of ModelChoiceField initialization.

        Most arguments given to ModelChoiceField can be set, with the exception
        of queryset, to_field_name, limit_choices_to and widget.

        :param scope: Scope of users to include:
                      "all"/"project"/"project_exclude"
        :param project: Project object or project UUID string (optional)
        :param exclude: List of User objects or User UUIDs to exclude
        :param forward: Parameters to forward to autocomplete view (optional)
        :param url: Autocomplete ajax class override (optional)
        :param widget_class: Widget class override (optional)
        """
        widget = get_user_widget(
            scope, project, exclude, forward, url, widget_class
        )
        super().__init__(
            User.objects.all(),
            to_field_name='sodar_uuid',
            limit_choices_to=None,
            widget=widget,
            *args,
            **kwargs
        )


# Project form -----------------------------------------------------------------


class ProjectForm(SODARModelForm):
    """Form for Project creation/updating"""

    # Set up owner field
    owner = SODARUserChoiceField(label='Owner', help_text='Owner')

    class Meta:
        model = Project
        fields = [
            'title',
            'type',
            'parent',
            'owner',
            'description',
            'readme',
            'public_guest_access',
        ]

    @classmethod
    def _get_parent_choices(cls, instance, user):
        """
        Return valid choices of parent categories for moving a project.

        :param instance: Project instance being updated
        :param user: Request user
        :return: List of tuples
        """
        ret = []
        # Add empty choice (root) in cases where valid
        if (
            user.is_superuser
            or not instance.parent
            or (
                instance.type == PROJECT_TYPE_PROJECT
                and getattr(settings, 'PROJECTROLES_DISABLE_CATEGORIES', False)
            )
        ):
            ret.append((None, EMPTY_CHOICE_LABEL))

        categories = Project.objects.filter(type=PROJECT_TYPE_CATEGORY).exclude(
            pk=instance.pk
        )

        if not user.is_superuser:
            categories = categories.filter(
                roles__in=RoleAssignment.objects.filter(
                    user=user,
                    role__name__in=[
                        PROJECT_ROLE_OWNER,
                        PROJECT_ROLE_DELEGATE,
                        PROJECT_ROLE_CONTRIBUTOR,
                    ],
                )
            )
            # Add categories with inherited ownership
            role_owner = Role.objects.filter(name=PROJECT_ROLE_OWNER).first()
            owned_cats = [
                r.project.full_title + CAT_DELIMITER
                for r in RoleAssignment.objects.filter(
                    project__type=PROJECT_TYPE_CATEGORY,
                    user=user,
                    role=role_owner,
                )
            ]
            other_categories = Project.objects.filter(
                type=PROJECT_TYPE_CATEGORY
            ).exclude(pk__in=categories)
            categories = list(categories)
            for c in other_categories:
                if any(c.full_title.startswith(t) for t in owned_cats):
                    categories.append(c)

        # If instance is category, exclude children
        if instance.type == PROJECT_TYPE_CATEGORY:
            categories = [
                c
                for c in categories
                if (
                    c.parent != instance
                    and c not in instance.get_children(flat=True)
                )
            ]

        # FIX for #558: Ensure current parent is in choices
        if (
            categories
            and not user.is_superuser
            and instance.parent
            and instance.parent not in categories
        ):
            categories.append(instance.parent)

        ret += [(c.sodar_uuid, c.full_title) for c in categories]
        return sorted(ret, key=lambda x: x[1])

    def _set_app_setting_widget(self, app_name, s_field, s_key, s_val):
        """Internal helper for setting app setting widget and value"""
        s_widget_attrs = s_val.get('widget_attrs') or {}
        if 'placeholder' in s_val:
            s_widget_attrs['placeholder'] = s_val.get('placeholder')
        setting_kwargs = {
            'required': False,
            'label': s_val.get('label') or '{}.{}'.format(app_name, s_key),
            'help_text': s_val['description'],
        }
        if s_val['type'] == 'JSON':
            # NOTE: Attrs MUST be supplied here (#404)
            if 'class' in s_widget_attrs:
                s_widget_attrs['class'] += ' sodar-json-input'
            else:
                s_widget_attrs['class'] = 'sodar-json-input'
            self.fields[s_field] = forms.CharField(
                widget=forms.Textarea(attrs=s_widget_attrs), **setting_kwargs
            )
            if self.instance.pk:
                json_data = self.app_settings.get_app_setting(
                    app_name=app_name,
                    setting_name=s_key,
                    project=self.instance,
                )
            else:
                json_data = self.app_settings.get_default_setting(
                    app_name=app_name, setting_name=s_key
                )
            self.initial[s_field] = json.dumps(json_data)
        else:
            if s_val.get('options'):
                self.fields[s_field] = forms.ChoiceField(
                    choices=[
                        (int(option), int(option))
                        if s_val['type'] == 'INTEGER'
                        else (option, option)
                        for option in s_val.get('options')
                    ],
                    **setting_kwargs
                )
            elif s_val['type'] == 'STRING':
                self.fields[s_field] = forms.CharField(
                    max_length=APP_SETTING_VAL_MAXLENGTH,
                    widget=forms.TextInput(attrs=s_widget_attrs),
                    **setting_kwargs
                )
            elif s_val['type'] == 'INTEGER':
                self.fields[s_field] = forms.IntegerField(
                    widget=forms.NumberInput(attrs=s_widget_attrs),
                    **setting_kwargs
                )
            elif s_val['type'] == 'BOOLEAN':
                self.fields[s_field] = forms.BooleanField(**setting_kwargs)

            # Add optional attributes from plugin (#404)
            # NOTE: Experimental! Use at your own risk!
            self.fields[s_field].widget.attrs.update(s_widget_attrs)

            # Set initial value
            if self.instance.pk:
                self.initial[s_field] = self.app_settings.get_app_setting(
                    app_name=app_name,
                    setting_name=s_key,
                    project=self.instance,
                )
            else:
                self.initial[s_field] = self.app_settings.get_default_setting(
                    app_name=app_name, setting_name=s_key
                )

    def _set_app_setting_notes(self, s_field, s_val):
        """Internal helper for setting app setting label notes"""
        if s_val.get('user_modifiable') is False:
            self.fields[s_field].label += ' [HIDDEN]'
            self.fields[s_field].help_text += ' [HIDDEN FROM USERS]'
        if s_val.get('local', APP_SETTING_LOCAL_DEFAULT) is False:
            if self.instance.is_remote():
                self.fields[s_field].label += ' [DISABLED]'
                self.fields[
                    s_field
                ].help_text += ' [Only editable on source site]'
                self.fields[s_field].disabled = True
            else:
                self.fields[
                    s_field
                ].help_text += ' [Not editable on target sites]'

    def _init_app_settings(self):
        # Set up setting query kwargs
        self.p_kwargs = {}
        if not self.current_user.is_superuser:
            self.p_kwargs['user_modifiable'] = True
        self.app_settings = AppSettingAPI()
        self.app_plugins = sorted(get_active_plugins(), key=lambda x: x.name)

        # plugin == 'None' refers to projectroles app
        for plugin in self.app_plugins + [None]:
            # Show non-modifiable settings to superusers
            if plugin:
                app_name = plugin.name
                p_settings = self.app_settings.get_setting_defs(
                    APP_SETTING_SCOPE_PROJECT, plugin=plugin, **self.p_kwargs
                )
            else:
                app_name = APP_NAME
                p_settings = self.app_settings.get_setting_defs(
                    APP_SETTING_SCOPE_PROJECT,
                    app_name=app_name,
                    **self.p_kwargs
                )
            for s_key, s_val in p_settings.items():
                s_field = 'settings.{}.{}'.format(app_name, s_key)
                # Set widget and value
                self._set_app_setting_widget(app_name, s_field, s_key, s_val)
                # Set label notes
                self._set_app_setting_notes(s_field, s_val)

    def __init__(self, project=None, current_user=None, *args, **kwargs):
        """Override for form initialization"""
        super().__init__(*args, **kwargs)
        disable_categories = getattr(
            settings, 'PROJECTROLES_DISABLE_CATEGORIES', False
        )
        # Get current user for checking permissions for form items
        if current_user:
            self.current_user = current_user
        # Add settings fields
        self._init_app_settings()
        # Access parent project if present
        parent_project = None
        if project:
            parent_project = Project.objects.filter(sodar_uuid=project).first()

        # Update help texts to match DISPLAY_NAMES
        self.fields['title'].help_text = 'Title'
        self.fields['type'].help_text = 'Type of container ({} or {})'.format(
            get_display_name(PROJECT_TYPE_CATEGORY),
            get_display_name(PROJECT_TYPE_PROJECT),
        )
        self.fields['type'].choices = PROJECT_TYPE_CHOICES
        self.fields['parent'].help_text = 'Parent {} if nested'.format(
            get_display_name(PROJECT_TYPE_CATEGORY)
        )
        self.fields['description'].help_text = 'Short description'
        self.fields['readme'].help_text = 'README (optional, supports markdown)'

        ####################
        # Form modifications
        ####################

        # Modify ModelChoiceFields to use sodar_uuid
        self.fields['parent'].to_field_name = 'sodar_uuid'
        # Set readme widget with preview
        self.fields['readme'].widget = SODARPagedownWidget(
            attrs={'show_preview': True}
        )

        # Updating an existing project
        if self.instance.pk:
            # Set readme value as raw markdown
            self.initial['readme'] = self.instance.readme.raw
            # Hide project type selection
            self.fields['type'].widget = forms.HiddenInput()
            # Set hidden project field for autocomplete
            self.initial['project'] = self.instance
            # Set owner value
            self.initial['owner'] = self.instance.get_owner().user.sodar_uuid
            # Hide owner widget if updating (changed in member modification UI)
            self.fields['owner'].widget = forms.HiddenInput()

            # Set valid choices for parent
            if not disable_categories:
                self.fields['parent'].choices = self._get_parent_choices(
                    self.instance, self.current_user
                )
                # Hide widget if no valid choices are available
                if len(self.fields['parent'].choices) == 0:
                    self.fields['parent'].widget = forms.HiddenInput()
                # Set initial value for parent
                if self.instance.parent:
                    self.initial['parent'] = self.instance.parent.sodar_uuid
            else:  # Categories disabled
                # Hide parent selection
                self.fields['parent'].widget = forms.HiddenInput()

        # Project creation
        else:
            # Set hidden project field for autocomplete
            self.initial['project'] = None
            # Hide parent selection
            self.fields['parent'].widget = forms.HiddenInput()

            # Set owner
            if self.current_user.is_superuser and parent_project:
                self.initial['owner'] = parent_project.get_owner().user
            else:
                self.initial['owner'] = self.current_user
            # Hide owner select widget for regular users
            if not self.current_user.is_superuser:
                self.fields['owner'].widget = forms.HiddenInput()

            # Creating a subproject
            if parent_project:
                # Parent must be current parent
                self.initial['parent'] = parent_project.sodar_uuid
                # Set initial value for type
                self.initial['type'] = None
            # Creating a top level project
            else:
                # Force project type
                if disable_categories:
                    self.initial['type'] = PROJECT_TYPE_PROJECT
                else:
                    self.initial['type'] = PROJECT_TYPE_CATEGORY
                # Hide project type selection
                self.fields['type'].widget = forms.HiddenInput()
                # Set up parent field
                self.initial['parent'] = None

        if self.instance.is_remote():
            self.fields['title'].widget = forms.HiddenInput()
            self.fields['type'].widget = forms.HiddenInput()
            self.fields['parent'].widget = forms.HiddenInput()
            self.fields['description'].widget = forms.HiddenInput()
            self.fields['readme'].widget = forms.HiddenInput()

    def clean(self):
        """Function for custom form validation and cleanup"""
        instance_owner_as = self.instance.get_owner() if self.instance else None
        disable_categories = getattr(
            settings, 'PROJECTROLES_DISABLE_CATEGORIES', False
        )
        parent = self.cleaned_data.get('parent')

        # Check for category/project being placed in root
        if not parent and (not self.instance or self.instance.parent):
            if (
                self.cleaned_data.get('type') == PROJECT_TYPE_CATEGORY
                and not self.current_user.is_superuser
            ):
                self.add_error(
                    'parent',
                    'You do not have permission to place a {} under root'.format(
                        get_display_name(PROJECT_TYPE_CATEGORY)
                    ),
                )
            elif (
                self.cleaned_data.get('type') == PROJECT_TYPE_PROJECT
                and not disable_categories
            ):
                self.add_error(
                    'parent', 'Projects can not be placed under root'
                )

        # Ensure title does not match parent
        if parent and parent.title == self.cleaned_data.get('title'):
            self.add_error(
                'title',
                '{} and parent titles can not be equal'.format(
                    get_display_name(self.cleaned_data.get('type'), title=True)
                ),
            )

        # Ensure title is unique within parent
        existing_project = Project.objects.filter(
            parent=self.cleaned_data.get('parent'),
            title=self.cleaned_data.get('title'),
        ).first()

        if existing_project and (
            not self.instance or existing_project.pk != self.instance.pk
        ):
            self.add_error('title', 'Title must be unique within parent')

        # Ensure owner has been set
        if not self.cleaned_data.get('owner'):
            self.add_error(
                'owner',
                'Owner must be set for {}'.format(
                    get_display_name(self.cleaned_data.get('type'))
                ),
            )

        # Ensure owner is not changed on update (must use ownership transfer)
        if (
            instance_owner_as
            and self.cleaned_data.get('owner') != instance_owner_as.user
        ):
            self.add_error(
                'owner',
                'Owner update is not allowed in this form, use Ownership '
                'Transfer instead',
            )

        # Verify settings fields
        for plugin in self.app_plugins + [None]:
            if plugin:
                name = plugin.name
                p_settings = self.app_settings.get_setting_defs(
                    APP_SETTING_SCOPE_PROJECT, plugin=plugin, **self.p_kwargs
                )
            else:
                name = 'projectroles'
                p_settings = self.app_settings.get_setting_defs(
                    APP_SETTING_SCOPE_PROJECT, app_name=name, **self.p_kwargs
                )

            for s_key, s_val in p_settings.items():
                s_field = 'settings.{}.{}'.format(name, s_key)

                if s_val['type'] == 'JSON':
                    # for some reason, there is a distinct possibility, that the
                    # initial value has been discarded and we get '' as value.
                    # Seems to only happen in automated tests. Will catch that
                    # here.
                    if not self.cleaned_data.get(s_field):
                        self.cleaned_data[s_field] = '{}'
                    try:
                        self.cleaned_data[s_field] = json.loads(
                            self.cleaned_data.get(s_field)
                        )
                    except json.JSONDecodeError as err:
                        # TODO: Shouldn't we use add_error() instead?
                        raise forms.ValidationError(
                            'Couldn\'t encode JSON\n' + str(err)
                        )
                elif s_val['type'] == 'INTEGER':
                    # When the field is a select/dropdown the information of
                    # the datatype gets lost. We need to convert that here,
                    # otherwise subsequent checks will fail.
                    self.cleaned_data[s_field] = int(self.cleaned_data[s_field])

                if not self.app_settings.validate_setting(
                    setting_type=s_val['type'],
                    setting_value=self.cleaned_data.get(s_field),
                    setting_options=s_val.get('options'),
                ):
                    self.add_error(s_field, 'Invalid value')

        return self.cleaned_data


# RoleAssignment form ----------------------------------------------------------


class RoleAssignmentForm(SODARModelForm):
    """Form for editing Project role assignments"""

    class Meta:
        model = RoleAssignment
        fields = ['project', 'user', 'role']

    def __init__(self, project=None, current_user=None, *args, **kwargs):
        """Override for form initialization"""
        super().__init__(*args, **kwargs)
        # Get current user for checking permissions for form items
        if current_user:
            self.current_user = current_user
        # Get the project for which role is being assigned
        self.project = None

        if self.instance.pk:
            self.project = self.instance.project
        else:
            self.project = Project.objects.filter(sodar_uuid=project).first()

        ####################
        # Form modifications
        ####################

        # Modify project field to use sodar_uuid
        self.fields['project'].to_field_name = 'sodar_uuid'
        # Set up user field
        self.fields['user'] = SODARUserChoiceField(
            scope='project_exclude',
            project=self.project,
            forward=['role'],
            url=reverse('projectroles:ajax_autocomplete_user_redirect'),
            widget_class=SODARUserRedirectWidget,
        )
        # Limit role choices
        self.fields['role'].choices = get_role_choices(
            self.project, self.current_user
        )

        # Updating an existing assignment
        if self.instance.pk:
            # Set values
            self.initial['project'] = self.instance.project.sodar_uuid
            self.initial['user'] = self.instance.user.sodar_uuid
            # Hide project and user switching
            self.fields['project'].widget = forms.HiddenInput()
            self.fields['user'].widget = forms.HiddenInput()
            # Set initial role
            self.initial['role'] = self.instance.role

        # Creating a new assignment
        elif self.project:
            # Limit project choice to self.project, hide widget
            self.initial['project'] = self.project.sodar_uuid
            self.fields['project'].widget = forms.HiddenInput()
            self.fields['role'].initial = Role.objects.get(
                name=PROJECT_ROLE_GUEST
            ).pk

    def clean(self):
        """Function for custom form validation and cleanup"""
        role = self.cleaned_data.get('role')
        existing_as = RoleAssignment.objects.get_assignment(
            self.cleaned_data.get('user'), self.cleaned_data.get('project')
        )

        # Adding a new RoleAssignment
        if not self.instance.pk:
            # Make sure user doesn't already have role in project
            if existing_as:
                self.add_error(
                    'user',
                    'User {} already assigned as {}'.format(
                        existing_as.role.name,
                        get_user_display_name(self.cleaned_data.get('user')),
                    ),
                )

        # Updating a RoleAssignment
        else:
            # Ensure not setting existing role again
            if existing_as and existing_as.role == role:
                self.add_error('role', 'Role already assigned to user')

        # Delegate checks
        if role.name == PROJECT_ROLE_DELEGATE:
            del_limit = getattr(settings, 'PROJECTROLES_DELEGATE_LIMIT', 1)

            # Ensure current user has permission to set delegate
            if not self.current_user.has_perm(
                'projectroles.update_project_delegate', obj=self.project
            ):
                self.add_error(
                    'role', 'Insufficient permissions for altering delegate'
                )

            # Ensure delegate limit is not exceeded
            if (
                del_limit != 0
                and self.project.get_delegates(exclude_inherited=True).count()
                >= del_limit
            ):
                self.add_error(
                    'role',
                    'The limit ({}) of delegates for this {} has '
                    'already been reached.'.format(
                        del_limit, get_display_name(self.project.type)
                    ),
                )

        return self.cleaned_data


# Owner transfer form ----------------------------------------------------------


class RoleAssignmentOwnerTransferForm(SODARForm):
    """Form for transferring owner role assignment between users"""

    def __init__(self, project, current_user, current_owner, *args, **kwargs):
        """Override for form initialization"""
        super().__init__(*args, **kwargs)
        # Get current user for checking permissions for form items
        self.current_user = current_user
        # Get the project for which role is being assigned
        self.project = project
        self.current_owner = current_owner

        self.fields['new_owner'] = SODARUserChoiceField(
            label='New owner',
            help_text='Select a member of the {} to become owner.'.format(
                get_display_name(self.project.type)
            ),
            scope='project',
            project=self.project,
            exclude=[self.current_owner],
        )

        self.selectable_roles = get_role_choices(
            self.project, self.current_user
        )
        self.fields['old_owner_role'] = forms.ChoiceField(
            label='New role for {}'.format(self.current_owner.username),
            help_text='New role for the current owner. Select "Remove" in the '
            'member list to remove the user\'s membership.',
            choices=self.selectable_roles,
            initial=Role.objects.get(name=PROJECT_ROLE_CONTRIBUTOR).pk,
        )
        self.fields['project'] = forms.Field(
            widget=forms.HiddenInput(), initial=self.project.sodar_uuid
        )

    def clean_old_owner_role(self):
        role = next(
            (
                choice
                for choice in self.selectable_roles
                if choice[0] == int(self.cleaned_data['old_owner_role'])
            ),
            None,
        )

        try:
            role = Role.objects.get(name=role[1])
        except Role.DoesNotExist:
            raise forms.ValidationError('Selected role does not exist')

        if role.name == PROJECT_ROLE_DELEGATE:
            del_limit = getattr(settings, 'PROJECTROLES_DELEGATE_LIMIT', 1)

            # Ensure current user has permission to set delegate
            if not self.current_user.has_perm(
                'projectroles.update_project_delegate', obj=self.project
            ):
                raise forms.ValidationError(
                    'Insufficient permissions for assigning a delegate role'
                )

            # Ensure delegate limit is not exceeded
            new_owner_role = RoleAssignment.objects.filter(
                project=self.project, user=self.cleaned_data.get('new_owner')
            ).first()

            if (
                del_limit != 0
                and new_owner_role
                and new_owner_role.role.name != PROJECT_ROLE_DELEGATE
                and self.project.get_delegates(exclude_inherited=True).count()
                >= del_limit
            ):
                raise forms.ValidationError(
                    'The limit of delegates ({}) for this {} has already been '
                    'reached.'.format(
                        del_limit, get_display_name(self.project.type)
                    )
                )

        return role

    def clean_new_owner(self):
        user = self.cleaned_data['new_owner']
        if user == self.current_owner:
            raise forms.ValidationError(
                'The new owner shouldn\'t be the current owner.'
            )

        role_as = RoleAssignment.objects.get_assignment(user, self.project)
        inh_owners = [
            a.user for a in self.project.get_owners(inherited_only=True)
        ]

        if (role_as and role_as.project != self.project) or (
            not role_as and user not in inh_owners
        ):
            raise forms.ValidationError(
                'The new owner has no roles in the {}.'.get_display_name(
                    self.project.type
                )
            )

        return user


# ProjectInvite form -----------------------------------------------------------


class ProjectInviteForm(SODARModelForm):
    """Form for ProjectInvite modification"""

    class Meta:
        model = ProjectInvite
        fields = ['email', 'role', 'message']

    def __init__(
        self,
        project=None,
        current_user=None,
        mail=None,
        role=None,
        *args,
        **kwargs
    ):
        """Override for form initialization"""
        super().__init__(*args, **kwargs)
        # Get current user for checking permissions and saving issuer
        if current_user:
            self.current_user = current_user
        # in case it has been redirected from the RoleAssignment form
        if mail:
            self.fields['email'].initial = mail
        # Get the project for which invite is being sent
        self.project = Project.objects.filter(sodar_uuid=project).first()
        # Limit Role choices according to user permissions
        self.fields['role'].choices = get_role_choices(
            self.project, self.current_user, allow_delegate=True
        )

        if role:
            self.fields['role'].initial = role
        else:
            self.fields['role'].initial = Role.objects.get(
                name=PROJECT_ROLE_GUEST
            ).pk

        # Limit textarea height
        self.fields['message'].widget.attrs['rows'] = 4

    def clean(self):
        # Check if user email is already in users
        existing_user = User.objects.filter(
            email=self.cleaned_data.get('email')
        ).first()
        if existing_user:
            self.add_error(
                'email',
                'User "{}" already exists in the system with this email. '
                'Please use "Add Role" instead.'.format(existing_user.username),
            )

        # Check if user already has an invite in the project
        try:
            ProjectInvite.objects.get(
                project=self.project,
                email=self.cleaned_data.get('email'),
                active=True,
                date_expire__gt=timezone.now(),
            )
            self.add_error(
                'email',
                'There is already an active invite for email {} in {}'.format(
                    self.cleaned_data.get('email'), self.project.title
                ),
            )
        except ProjectInvite.DoesNotExist:
            pass

        # Delegate checks
        role = self.cleaned_data.get('role')
        if role.name == PROJECT_ROLE_DELEGATE:
            del_limit = getattr(settings, 'PROJECTROLES_DELEGATE_LIMIT', 1)
            # Ensure current user has permission to invite delegate
            if not self.current_user.has_perm(
                'projectroles.update_project_delegate', obj=self.project
            ):
                self.add_error(
                    'role', 'Insufficient permissions for inviting delegate.'
                )
            # Ensure delegate limit is not exceeded
            if (
                del_limit != 0
                and self.project.get_delegates(exclude_inherited=True).count()
                >= del_limit
            ):
                self.add_error(
                    'role',
                    'The delegate limit of {} for this {} has '
                    'already been reached.'.format(
                        del_limit, get_display_name(self.project.type)
                    ),
                )

        return self.cleaned_data

    def save(self, *args, **kwargs):
        """Override of form saving function"""
        obj = super().save(commit=False)
        obj.project = self.project
        obj.issuer = self.current_user
        obj.date_expire = timezone.now() + timezone.timedelta(
            days=settings.PROJECTROLES_INVITE_EXPIRY_DAYS
        )
        obj.secret = build_secret()
        obj.save()
        return obj


# RemoteSite form --------------------------------------------------------------


class RemoteSiteForm(SODARModelForm):
    """Form for RemoteSite creation/updating"""

    class Meta:
        model = RemoteSite
        fields = ['name', 'url', 'description', 'user_display', 'secret']

    def __init__(self, current_user=None, *args, **kwargs):
        """Override for form initialization"""
        super().__init__(*args, **kwargs)
        self.current_user = current_user

        # Default field modifications
        self.fields['description'].required = False
        self.fields['secret'].widget = forms.TextInput(
            attrs={'class': 'sodar-code-input'}
        )
        self.fields['description'].widget.attrs['rows'] = 4

        # Special cases for SOURCE
        if settings.PROJECTROLES_SITE_MODE == SITE_MODE_SOURCE:
            self.fields['secret'].widget.attrs['readonly'] = True
            self.fields['user_display'].widget = forms.CheckboxInput()
        elif settings.PROJECTROLES_SITE_MODE == SITE_MODE_TARGET:
            self.fields['user_display'].widget = forms.HiddenInput()
        self.fields['user_display'].initial = True

        # Creation
        if not self.instance.pk:
            # Generate secret token for target site
            if settings.PROJECTROLES_SITE_MODE == SITE_MODE_SOURCE:
                self.fields['secret'].initial = build_secret()
        # Updating
        else:
            pass

    def save(self, *args, **kwargs):
        """Override of form saving function"""
        obj = super().save(commit=False)
        if settings.PROJECTROLES_SITE_MODE == SITE_MODE_SOURCE:
            obj.mode = SITE_MODE_TARGET
        else:
            obj.mode = SITE_MODE_SOURCE
        obj.save()
        return obj


# Local user editing form ------------------------------------------------------


class LocalUserForm(SODARModelForm):
    """Form for local user creation and updating"""

    password_confirm = forms.CharField(
        label='Confirm password', widget=forms.PasswordInput()
    )

    class Meta:
        model = User
        fields = [
            'first_name',
            'last_name',
            'email',
            'username',
            'password',
            'password_confirm',
        ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['email'].widget.attrs['readonly'] = True
        self.fields['username'].widget.attrs['readonly'] = True
        self.fields['password'].widget = forms.PasswordInput()

    def save(self, commit=True):
        user = super().save(commit=False)
        user.set_password(self.cleaned_data['password'])
        if commit:
            user.save()
        return user

    def clean(self):
        cleaned_data = super().clean()
        if cleaned_data['password'] != cleaned_data['password_confirm']:
            self.add_error('password_confirm', 'Passwords didn\'t match!')
            self.add_error('password', 'Passwords didn\'t match!')
        return cleaned_data


# Helper functions -------------------------------------------------------------


def get_role_choices(
    project, current_user, allow_delegate=True, allow_owner=False
):
    """
    Return valid role choices according to permissions of current user.

    :param project: Project in which role will be assigned
    :param current_user: User for whom the form is displayed
    :param allow_delegate: Whether delegate setting should be allowed (bool)
    """
    # Owner cannot be changed in role assignment
    role_excludes = []
    if not allow_owner or not current_user.has_perm(
        'projectroles.update_project_owner', obj=project
    ):
        role_excludes.append(PROJECT_ROLE_OWNER)
    # Exclude delegate if not allowed or current user lacks perms
    if not allow_delegate or not current_user.has_perm(
        'projectroles.update_project_delegate', obj=project
    ):
        role_excludes.append(PROJECT_ROLE_DELEGATE)
    return [
        (role.pk, role.name)
        for role in Role.objects.exclude(name__in=role_excludes)
    ]



from projectroles.plugins import SiteAppPluginPoint


class ProjectAppPlugin(SiteAppPluginPoint):
    """Plugin for registering app with Projectroles"""

    name = 'token'

    title = 'API Tokens'

    #: Iconify icon
    icon = 'mdi:key-chain-variant'

    entry_point_url_id = 'tokens:list'

    description = 'API Token Management'

    #: Required permission for accessing the app
    app_permission = 'tokens.access'
